#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Include a commitment inside an elliptic curve DSA/SSA signature.

Let c be the commitment value and R a curve point, then

    e = hash(R||c)

is a commitment operation.

When signing, an ephemeral secret key k is generated and its
corresponding curve point R = kG is used. Here, instead of
using (k, R), compute the commitment to c

    e = hash(R||c),

tweak k with e and consequently substistute R with W = (k+e)G = R+eG,
the proceed signing in the standard way, using (k+e, W).

When the committer/signer will reveal R and c,
the verifier will check that

    W.x = (R+eG).x

with e = hash(R||c)) and W.x being known from the signature.
"""

from typing import Optional, Tuple
from hashlib import sha256

from .curve import Curve, Point
from .curves import secp256k1
from .curvemult import mult
from .utils import int_from_bits, point_from_octets, octets_from_point, HashF
from .rfc6979 import rfc6979
from . import dsa
from . import ssa

# commitment receipt
Receipt = Tuple[int, Point]


def _tweak(c: bytes, k: int,
           ec: Curve = secp256k1, hf: HashF = sha256) -> Tuple[Point, int]:
    """Tweak kG with hash(kG||c).

    Return:
    - the point kG to tweak
    - tweaked private key k + hash(kG||c)
    """
    R = mult(k, ec.G, ec)
    e = hf(octets_from_point(R, True, ec) + c).digest()
    e = int.from_bytes(e, byteorder='big')
    return R, (e + k) % ec.n


def ecdsa_commit_sign(c: bytes, m: bytes, prvkey: int, k: Optional[int] = None,
                      ec: Curve = secp256k1, hf: HashF = sha256) -> Tuple[dsa.ECDS, Receipt]:
    """Include a commitment c inside an ECDSA signature."""

    if k is None:
        k = rfc6979(hf(m).digest(), prvkey, ec, hf)

    ch = hf(c).digest()

    # commit
    R, new_k = _tweak(ch, k, ec, hf)
    # sign
    sig = dsa.sign(m, prvkey, new_k, ec, hf)
    # commit receipt
    receipt = sig[0], R
    return sig, receipt


def ecssa_commit_sign(c: bytes, m: bytes, prvkey: int, k: Optional[int] = None,
                      ec: Curve = secp256k1, hf: HashF = sha256) -> Tuple[ssa.ECSS, Receipt]:
    """Include a commitment c inside an ECSSA signature."""

    if k is None:
        k = rfc6979(m, prvkey, ec, hf)

    ch = hf(c).digest()

    # commit
    R, new_k = _tweak(ch, k, ec, hf)
    # sign
    sig = ssa.sign(m, prvkey, new_k, ec, hf)
    # commit receipt
    receipt = sig[0], R
    return sig, receipt

# FIXME: have create_commit instead of commit_sign


def verify_commit(c: bytes, receipt: Receipt,
                  ec: Curve = secp256k1, hf: HashF = sha256) -> bool:
    """Open the commitment c inside an EC DSA/SSA signature."""

    # FIXME: verify the signature

    w, R = receipt
    # w in [1..n-1] dsa
    # w in [1..p-1] ssa
    # different verify functions?

    # verify R is a good point?

    ch = hf(c).digest()
    e = hf(octets_from_point(R, True, ec) + ch).digest()
    e = int_from_bits(e, ec)
    W = ec.add(R, mult(e, ec.G, ec))
    # different verify functions?
    # return w == W[0] # ECSS
    return w == W[0] % ec.n  # ECDS, FIXME: ECSSA



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Bitcoin Script.

https://en.bitcoin.it/wiki/Script

Scripts are represented by Iterable[Token] (usually List[Token]),
where Token = Union[int, str, bytes]:

* int [-1, 16] are shorcuts for 'OP_1NEGATE', 'OP_0' - 'OP_16'
* str are for opcodes (e.g. 'OP_HASH160') or hexstring data
* bytes are for data (but integers are often casted to int)
"""

OP_CODES = {
    # Constants
    'OP_0'                  : b'\x00',
    'OP_FALSE'              : b'\x00',
    'OP_PUSHDATA1'          : b'\x4c',
    'OP_PUSHDATA2'          : b'\x4d',
    'OP_PUSHDATA4'          : b'\x4e',
    'OP_1NEGATE'            : b'\x4f',
    'OP_1'                  : b'\x51',
    'OP_TRUE'               : b'\x51',
    'OP_2'                  : b'\x52',
    'OP_3'                  : b'\x53',
    'OP_4'                  : b'\x54',
    'OP_5'                  : b'\x55',
    'OP_6'                  : b'\x56',
    'OP_7'                  : b'\x57',
    'OP_8'                  : b'\x58',
    'OP_9'                  : b'\x59',
    'OP_10'                 : b'\x5a',
    'OP_11'                 : b'\x5b',
    'OP_12'                 : b'\x5c',
    'OP_13'                 : b'\x5d',
    'OP_14'                 : b'\x5e',
    'OP_15'                 : b'\x5f',
    'OP_16'                 : b'\x60',

    # Flow control
    'OP_NOP'                : b'\x61',
    'OP_IF'                 : b'\x63',
    'OP_NOTIF'              : b'\x64',
    'OP_ELSE'               : b'\x67',
    'OP_ENDIF'              : b'\x68',
    'OP_VERIFY'             : b'\x69',
    'OP_RETURN'             : b'\x6a',

    # Stack
    'OP_TOALTSTACK'         : b'\x6b',
    'OP_FROMALTSTACK'       : b'\x6c',
    'OP_IFDUP'              : b'\x73',
    'OP_DEPTH'              : b'\x74',
    'OP_DROP'               : b'\x75',
    'OP_DUP'                : b'\x76',
    'OP_NIP'                : b'\x77',
    'OP_OVER'               : b'\x78',
    'OP_PICK'               : b'\x79',
    'OP_ROLL'               : b'\x7a',
    'OP_ROT'                : b'\x7b',
    'OP_SWAP'               : b'\x7c',
    'OP_TUCK'               : b'\x7d',
    'OP_2DROP'              : b'\x6d',
    'OP_2DUP'               : b'\x6e',
    'OP_3DUP'               : b'\x6f',
    'OP_2OVER'              : b'\x70',
    'OP_2ROT'               : b'\x71',
    'OP_2SWAP'              : b'\x72',

    # Splice
    'OP_SIZE'               : b'\x82',

    # Bitwise logic
    'OP_EQUAL'              : b'\x87',
    'OP_EQUALVERIFY'        : b'\x88',

    # Arithmetic
    'OP_1ADD'               : b'\x8b',
    'OP_1SUB'               : b'\x8c',
    'OP_NEGATE'             : b'\x8f',
    'OP_ABS'                : b'\x90',
    'OP_NOT'                : b'\x91',
    'OP_0NOTEQUAL'          : b'\x92',
    'OP_ADD'                : b'\x93',
    'OP_SUB'                : b'\x94',
    'OP_BOOLAND'            : b'\x9a',
    'OP_BOOLOR'             : b'\x9b',
    'OP_NUMEQUAL'           : b'\x9c',
    'OP_NUMEQUALVERIFY'     : b'\x9d',
    'OP_NUMNOTEQUAL'        : b'\x9e',
    'OP_LESSTHAN'           : b'\x9f',
    'OP_GREATERTHAN'        : b'\xa0',
    'OP_LESSTHANOREQUAL'    : b'\xa1',
    'OP_GREATERTHANOREQUAL' : b'\xa2',
    'OP_MIN'                : b'\xa3',
    'OP_MAX'                : b'\xa4',
    'OP_WITHIN'             : b'\xa5',

    # Crypto
    'OP_RIPEMD160'          : b'\xa6',
    'OP_SHA1'               : b'\xa7',
    'OP_SHA256'             : b'\xa8',
    'OP_HASH160'            : b'\xa9',
    'OP_HASH256'            : b'\xaa',
    'OP_CODESEPARATOR'      : b'\xab',
    'OP_CHECKSIG'           : b'\xac',
    'OP_CHECKSIGVERIFY'     : b'\xad',
    'OP_CHECKMULTISIG'      : b'\xae',
    'OP_CHECKMULTISIGVERIFY': b'\xaf',

    # Locktime
    'OP_NOP2'               : b'\xb1',
    'OP_CHECKLOCKTIMEVERIFY': b'\xb1',
    'OP_NOP3'               : b'\xb2',
    'OP_CHECKSEQUENCEVERIFY': b'\xb2',

    # Reserved words
    'OP_NOP1'               : b'\xb0',
    'OP_NOP4'               : b'\xb3',
    'OP_NOP5'               : b'\xb4',
    'OP_NOP6'               : b'\xb5',
    'OP_NOP7'               : b'\xb6',
    'OP_NOP8'               : b'\xb7',
    'OP_NOP9'               : b'\xb8',
    'OP_NOP10'              : b'\xb9'
}

OP_CODE_NAMES = {
      0: 'OP_0',
     76: 'OP_PUSHDATA1',
     77: 'OP_PUSHDATA2',
     78: 'OP_PUSHDATA4',
     79: 'OP_1NEGATE',
     81: 'OP_1',
     82: 'OP_2',
     83: 'OP_3',
     84: 'OP_4',
     85: 'OP_5',
     86: 'OP_6',
     87: 'OP_7',
     88: 'OP_8',
     89: 'OP_9',
     90: 'OP_10',
     91: 'OP_11',
     92: 'OP_12',
     93: 'OP_13',
     94: 'OP_14',
     95: 'OP_15',
     96: 'OP_16',
     97: 'OP_NOP',
     99: 'OP_IF',
    100: 'OP_NOTIF',
    103: 'OP_ELSE',
    104: 'OP_ENDIF',
    105: 'OP_VERIFY',
    106: 'OP_RETURN',
    107: 'OP_TOALTSTACK',
    108: 'OP_FROMALTSTACK',
    109: 'OP_2DROP',
    110: 'OP_2DUP',
    111: 'OP_3DUP',
    112: 'OP_2OVER',
    113: 'OP_2ROT',
    114: 'OP_2SWAP',
    115: 'OP_IFDUP',
    116: 'OP_DEPTH',
    117: 'OP_DROP',
    118: 'OP_DUP',
    119: 'OP_NIP',
    120: 'OP_OVER',
    121: 'OP_PICK',
    122: 'OP_ROLL',
    123: 'OP_ROT',
    124: 'OP_SWAP',
    125: 'OP_TUCK',
    130: 'OP_SIZE',
    135: 'OP_EQUAL',
    136: 'OP_EQUALVERIFY',
    139: 'OP_1ADD',
    140: 'OP_1SUB',
    143: 'OP_NEGATE',
    144: 'OP_ABS',
    145: 'OP_NOT',
    146: 'OP_0NOTEQUAL',
    147: 'OP_ADD',
    148: 'OP_SUB',
    154: 'OP_BOOLAND',
    155: 'OP_BOOLOR',
    156: 'OP_NUMEQUAL',
    157: 'OP_NUMEQUALVERIFY',
    158: 'OP_NUMNOTEQUAL',
    159: 'OP_LESSTHAN',
    160: 'OP_GREATERTHAN',
    161: 'OP_LESSTHANOREQUAL',
    162: 'OP_GREATERTHANOREQUAL',
    163: 'OP_MIN',
    164: 'OP_MAX',
    165: 'OP_WITHIN',
    166: 'OP_RIPEMD160',
    167: 'OP_SHA1',
    168: 'OP_SHA256',
    169: 'OP_HASH160',
    170: 'OP_HASH256',
    171: 'OP_CODESEPARATOR',
    172: 'OP_CHECKSIG',
    173: 'OP_CHECKSIGVERIFY',
    174: 'OP_CHECKMULTISIG',
    175: 'OP_CHECKMULTISIGVERIFY',
    176: 'OP_NOP1',
    177: 'OP_CHECKLOCKTIMEVERIFY',
    178: 'OP_CHECKSEQUENCEVERIFY',
    179: 'OP_NOP4',
    180: 'OP_NOP5',
    181: 'OP_NOP6',
    182: 'OP_NOP7',
    183: 'OP_NOP8',
    184: 'OP_NOP9',
    185: 'OP_NOP10',
}

from io import BytesIO
from typing import BinaryIO, Iterable, List, Union

from . import varint
from .utils import Octets, bytes_from_hexstring

# the integers [0-16] are shorcuts for 'OP_0'-'OP_16'
# the integer -1 is a shorcut for 'OP_1NEGATE'
# other integers are bytes encoded (require push operation) 
# str are for opcodes (e.g. 'OP_HASH160') or hexstring data to be pushed
# bytes is for data to be pushed
Token = Union[int, str, bytes]
# TODO: use Script as input and serialize Iterable[Token] to bytes on the fly
Script = Union[bytes, Iterable[Token]]


def _op_pushdata(data: bytes) -> bytes:
    """Convert to canonical push: OP_PUSHDATA (if needed) | length | data.

    According to standardness rules (BIP-62) the
    minimum possible PUSHDATA operator must be used.
    Byte vectors on the stack are not allowed to be more than 520 bytes long.
    """

    r = b''
    length = len(data)
    if length < 75:     # 1-byte-length
        r += length.to_bytes(1, byteorder='little')
    elif length < 256:  # OP_PUSHDATA1 | 1-byte-length
        r += OP_CODES['OP_PUSHDATA1']
        r += length.to_bytes(1, byteorder='little')
    elif length < 521:  # OP_PUSHDATA2 | 2-byte-length
        r += OP_CODES['OP_PUSHDATA2']
        r += length.to_bytes(2, byteorder='little')
    else:
        # because of the 520 bytes limit
        # there is no need to use OP_PUSHDATA4
        # r += OP_CODES['OP_PUSHDATA4']
        # r += length.to_bytes(4, byteorder='little')
        raise ValueError(f"Script: Cannot push {length} bytes on the stack")
    r += data
    return r


def encode(script: Iterable[Token]) -> bytes:
    r = b''
    for token in script:
        if isinstance(token, int):
            # Short 1-byte opcodes exist
            # to push numbers in [-1, 16]
            if token >= 0 and token <= 16:
                r += OP_CODES['OP_' + str(token)]
            elif token == -1:
                r += OP_CODES['OP_1NEGATE']
            # Pushing any other number requires an
            # explicit push operation of its bytes encoding
            # FIXME: negative numbers?
            else:
                nbytes = (token.bit_length() + 7) // 8
                data = token.to_bytes(nbytes, byteorder='little')
                r += _op_pushdata(data)
        elif isinstance(token, str):
            token = token.strip()
            token = token.upper()
            if token in OP_CODES:
                r += OP_CODES[token]
            else:
                try:
                   token = bytes.fromhex(token)
                except Exception:
                    raise ValueError(f"Script: invalid {token!r} opcode")
                r += _op_pushdata(token)
        elif isinstance(token, bytes):
                r += _op_pushdata(token)
        else:
            raise ValueError(f"Script: unmanaged {type(token)} token type")
    return r


def decode(script: Octets) -> List[Token]:

    script = bytes_from_hexstring(script)

    # initialize the result list
    r: List[Union[str, int, bytes]] = []

    length = len(script)
    s = BytesIO(script)  # TODO: avoid unnecessary streamification
    counter = 0
    while counter < length:
        # get one byte and convert it to an integer
        i = s.read(1)[0]
        counter += 1
        if i == 0:
            # numeric value 0 (OP_0)
            r.append(i)
        elif i == 79:
            # numeric value -1 (OP_1NEGATE)
            r.append(-1)
        elif i > 80 and i < 97:
            # numeric values 1-16 (OP_1-OP_16)
            r.append(i-80)
        elif i < 76:
            # 1-byte-data-length | data
            data = s.read(i)
            counter += i
            n = int.from_bytes(data, byteorder='little')
            if n <= 0xffffffff:
                # probably a number, it won't hurt to consider it a number
                r.append(n)
            else:
                r.append(data.hex())
        elif i == 76:
            # OP_PUSHDATA1 | 1-byte-data-length | data
            data_length = int.from_bytes(s.read(1), byteorder='little')
            data = s.read(data_length)
            counter += 1 + data_length
            r.append(data.hex())
        elif i == 77:
            # OP_PUSHDATA2 | 2-byte-data-length | data
            data_length = int.from_bytes(s.read(2), byteorder='little')
            data = s.read(data_length)
            counter += 2 + data_length
            r.append(data.hex())
        elif i == 78:
            # OP_PUSHDATA4 | 4-byte-data-length | data
            data_length = int.from_bytes(s.read(4), byteorder='little')
            data = s.read(data_length)
            counter += 4 + data_length
            r.append(data.hex())
        else:
            # OP_CODE
            r.append(OP_CODE_NAMES[i])

    return r


def serialize(script: Iterable[Token]) -> bytes:
    r = encode(script)
    # prepend length as varint
    length  = len(r)
    return varint.encode(length) + r


def parse(stream: Union[BinaryIO, Octets]) -> List[Token]:

    stream = bytes_from_hexstring(stream)

    if isinstance(stream, bytes):
        stream = BytesIO(stream)

    length = varint.decode(stream)
    script = stream.read(length)
    return decode(script)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Class for word-lists to be used in entropy/mnemonic conversions."""

import math
from os import path
from typing import List

from btclib.utils import ensure_is_power_of_two

WordList = List[str]


class WordLists:
    """Class for word-lists to be used in entropy/mnemonic conversions.

    Word-lists are from:

    * *en*: https://github.com/bitcoin/bips/blob/master/bip-0039/english.txt
    * *it*: https://github.com/bitcoin/bips/blob/master/bip-0039/italian.txt

    More word-lists can be added using the load_lang method.

    Word-lists are loaded only if needed and read only once from disk.
    """

    def __init__(self) -> None:

        path_to_filename = path.join(path.dirname(__file__), "dictdata")
        self.language_files = {
            'en': path.join(path_to_filename, 'english.txt'),
            'it': path.join(path_to_filename, 'italian.txt')
        }
        self.languages = list(self.language_files)

        # create dictionaries where each language has empty word-list
        wordlists: List[List[str]] = [[] for _ in self.languages]
        self._wordlist = dict(zip(self.languages, wordlists))

        zeros = len(self.languages)*[0]
        self._bits_per_word = dict(zip(self.languages, zeros))
        self._language_length = dict(zip(self.languages, zeros))

    def load_lang(self, lang: str, filename: str = None) -> None:
        """Load/add a language word-list if not loaded/added yet.

        The language file has to be provided for adding new languages
        beyond those already provided. 
        """

        # a new language, unknown before
        if lang not in self.languages:
            if filename is None:
                raise ValueError(f"missing file for language '{lang}'")
            else:
                # initialize the new language
                self.languages.append(lang)
                self.language_files[lang] = filename
                self._wordlist[lang] = []
                self._bits_per_word[lang] = 0
                self._language_length[lang] = 0

        # language has not been loaded yet
        if self._language_length[lang] == 0:
            with open(self.language_files[lang], 'r') as f:
                lines = f.readlines()
            f.closed

            nwords = len(lines)
            ensure_is_power_of_two(nwords, "wordlist length")

            self._bits_per_word[lang] = int(math.log(nwords, 2))
            self._language_length[lang] = nwords
            # clean up and normalization are missing, but removal of \n
            self._wordlist[lang] = [line[:-1] for line in lines]

    def bits_per_word(self, lang: str) -> int:
        """Return the number of bits represented by a single word.

        Return the number of bits represented by a single word for
        the given language.
        """
        self.load_lang(lang)
        return self._bits_per_word[lang]

    def wordlist(self, lang: str) -> WordList:
        """Return the language word-list."""

        self.load_lang(lang)
        return self._wordlist[lang]

    def language_length(self, lang: str) -> int:
        """Return the number of words in the language word-list."""

        self.load_lang(lang)
        return self._language_length[lang]


# singleton
_wordlists = WordLists()



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Elliptic curve class and functions."""

import heapq
from typing import Sequence, List

from .curve import Curve, Point, _JacPoint, _jac_from_aff
from .curves import secp256k1

def mult(m: int, Q: Point = None, ec: Curve = secp256k1) -> Point:
    # this function is used by the Curve class; it might be a method...
    # but it does not need to
    if Q is None:
        QJ = ec.GJ
    else:
        ec.require_on_curve(Q)
        QJ = _jac_from_aff(Q)
    R = _mult_jac(m, QJ, ec)
    return ec._aff_from_jac(R)

def _mult_jac(m: int, Q: _JacPoint, ec: Curve = secp256k1) -> _JacPoint:
    # double & add in Jacobian coordinates, using binary decomposition of m
    # Point is assumed to be on curve

    m %= ec.n
    if m == 0 or Q[2] == 0:        # Infinity point in affine coordinates
        return 1, 1, 0             # return Infinity point
    R = 1, 1, 0                    # initialize as infinity point
    while m > 0:                   # use binary representation of m
        if m & 1:                  # if least significant bit is 1
            R = ec._add_jac(R, Q)  # then add current Q
        m = m >> 1                 # remove the bit just accounted for
        Q = ec._add_jac(Q, Q)      # double Q for next step
    return R


def double_mult(u: int, H: Point, v: int, Q: Point = None,
                ec: Curve = secp256k1) -> Point:
    """Shamir trick for efficient computation of u*H + v*Q"""

    ec.require_on_curve(H)
    HJ = _jac_from_aff(H)

    if Q is None:
        QJ = ec.GJ
    else:
        ec.require_on_curve(Q)
        QJ = _jac_from_aff(Q)

    R = _double_mult(u, HJ, v, QJ, ec)

    return ec._aff_from_jac(R)


def _double_mult(u: int, HJ: _JacPoint, v: int, QJ: _JacPoint,
                 ec: Curve = secp256k1) -> _JacPoint:

    u %= ec.n
    if u == 0 or HJ[2] == 0:
        return _mult_jac(v, QJ, ec)

    v %= ec.n
    if v == 0 or QJ[2] == 0:
        return _mult_jac(u, HJ, ec)

    R = 1, 1, 0  # initialize as infinity point
    msb = max(u.bit_length(), v.bit_length())
    while msb > 0:
        if u >> (msb - 1):  # checking msb
            R = ec._add_jac(R, HJ)
            u -= pow(2, u.bit_length() - 1)
        if v >> (msb - 1):  # checking msb
            R = ec._add_jac(R, QJ)
            v -= pow(2, v.bit_length() - 1)
        if msb > 1:
            R = ec._add_jac(R, R)
        msb -= 1

    return R


def multi_mult(scalars: Sequence[int], Points: Sequence[Point],
               ec: Curve = secp256k1) -> Point:
    """Return the multi scalar multiplication u1*Q1 + ... + un*Qn.

    Use Bos-Coster's algorithm for efficient computation;
    the input points must be on the curve.
    """

    if len(scalars) != len(Points):
        errMsg = f"mismatch between scalar length ({len(scalars)}) and "
        errMsg += f"Points length ({len(Points)})"
        raise ValueError(errMsg)

    JPoints: List[_JacPoint] = list()
    for P in Points:
        ec.require_on_curve(P)
        JPoints.append(_jac_from_aff(P))

    R = _multi_mult(scalars, JPoints, ec)

    return ec._aff_from_jac(R)


def _multi_mult(scalars: Sequence[int], JPoints: Sequence[_JacPoint],
                ec: Curve = secp256k1) -> _JacPoint:
    # source: https://cr.yp.to/badbatch/boscoster2.py

    x = list(zip([-n for n in scalars], JPoints))
    heapq.heapify(x)
    while len(x) > 1:
        np1 = heapq.heappop(x)
        np2 = heapq.heappop(x)
        n1, p1 = -np1[0], np1[1]
        n2, p2 = -np2[0], np2[1]
        p2 = ec._add_jac(p1, p2)
        n1 -= n2
        if n1 > 0:
            heapq.heappush(x, (-n1, p1))
        heapq.heappush(x, (-n2, p2))
    np1 = heapq.heappop(x)
    n1, p1 = -np1[0], np1[1]
    return _mult_jac(n1, p1, ec)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

""" Base58 address functions.

Base58 encoding of public keys and scripts as addresses.
"""

from typing import List, Tuple, Union

from .base58 import b58decode, b58encode
from .script import Script, Token
from .utils import (Octets, bytes_from_hexstring, hash160, int_from_octets,
                    octets_from_point)

_NETWORKS = ['mainnet', 'testnet', 'regtest']
_P2PKH_PREFIXES = [
    b'\x00',  # address starts with 1
    b'\x6f',  # address starts with {m, n}
    b'\x6f'   # address starts with {m, n}
]
_P2SH_PREFIXES = [
    b'\x05',  # address starts with 3
    b'\xc4',  # address starts with 2
    b'\xc4',  # address starts with 2
]


def _p2pkh_address(h160: bytes, network: str = 'mainnet') -> bytes:
    """Return the p2pkh address corresponding to a public key."""

    payload = _P2PKH_PREFIXES[_NETWORKS.index(network)]
    payload += h160
    return b58encode(payload)


def p2pkh_address(pubkey: Octets, network: str = 'mainnet') -> bytes:
    """Return the p2pkh address corresponding to a public key."""

    pubkey = bytes_from_hexstring(pubkey)
    if len(pubkey) not in (33, 65):
        raise ValueError(f"Invalid SEC pubkey length: {len(pubkey)}-bytes")
    h160 = hash160(pubkey)
    return _p2pkh_address(h160, network)


def _p2sh_address(h160: bytes, network: str = 'mainnet') -> bytes:
    """Return p2sh address."""

    payload = _P2SH_PREFIXES[_NETWORKS.index(network)]
    payload += h160
    return b58encode(payload)

def p2sh_address(script: Octets, network: str = 'mainnet') -> bytes:
    """Return p2sh address."""

    h160 = hash160(script)
    return _p2sh_address(h160, network)


def h160_from_base58_address(address: Union[str, bytes]) -> Tuple[str, bool, bytes]:

    if isinstance(address, str):
        address = address.strip()
    payload = b58decode(address, 21)
    prefix = payload[0:1]
    if prefix in _P2PKH_PREFIXES:
        i = _P2PKH_PREFIXES.index(prefix)
        is_script_hash = False
    elif prefix in _P2SH_PREFIXES:
        i = _P2SH_PREFIXES.index(prefix)
        is_script_hash = True
    else:
        raise ValueError(f"Invalid base58 address prefix {prefix!r}")

    return _NETWORKS[i], is_script_hash, payload[1:]



# Copyright (c) 2017 Pieter Wuille
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Copyright (C) 2019-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.


"""SegWit address functions.

Some of these functions were originally from
https://github.com/sipa/bech32/tree/master/ref/python,
with the following modifications:

* moved bech32 stuff into bech32.py
* type annotated python3
* avoided returning None or (None, None), throwing ValueError instead
* detailed error messages and exteded safety checks
* check that bech32 addresses are not longer than 90 characters
  (as this is not enforced by bech32._encode anymore)
"""


from typing import Iterable, List, Tuple, Union

from .bech32 import b32decode, b32encode
from .address import p2sh_address
from .script import Token, encode
from .utils import Octets, bytes_from_hexstring, hash160, sha256

WitnessProgram = Union[List[int], bytes]

_NETWORKS = ['mainnet', 'testnet', 'regtest']
_P2W_PREFIXES = ['bc', 'tb', 'bcrt']


def has_segwit_prefix(addr: Union[str, bytes]) -> bool:

    if isinstance(addr, bytes):
        str_addr = addr.decode()
    else:
        str_addr = addr.strip()
        str_addr = str_addr.lower()
    
    for prefix in _P2W_PREFIXES:
        if str_addr.startswith(prefix + '1'):
            return True

    return False


def _convertbits(data: Iterable[int], frombits: int,
                 tobits: int, pad: bool = True) -> List[int]:
    """General power-of-2 base conversion."""
    acc = 0
    bits = 0
    ret = []
    maxv = (1 << tobits) - 1
    max_acc = (1 << (frombits + tobits - 1)) - 1
    for value in data:
        if value < 0 or (value >> frombits):
            raise ValueError(f"invalid value {value}")
        acc = ((acc << frombits) | value) & max_acc
        bits += frombits
        while bits >= tobits:
            bits -= tobits
            ret.append((acc >> bits) & maxv)

    if pad:
        if bits:
            ret.append((acc << (tobits - bits)) & maxv)
    elif bits >= frombits or ((acc << (tobits - bits)) & maxv):
        raise ValueError("failure")

    return ret


def _check_witness(witvers: int, witprog: WitnessProgram):
    l = len(witprog)
    if witvers == 0:
        if l != 20 and l != 32:
            raise ValueError(f"witness program length ({l}) is not 20 or 32")
    elif witvers > 16 or witvers < 0:
        msg = f"witness version ({witvers}) not in [0, 16]"
        raise ValueError(msg)
    else:
        if l < 2 or l > 40:
            raise ValueError(f"witness program length ({l}) not in [2, 40]")


def _decode(address: Union[str, bytes]) -> Tuple[str, int, List[int]]:
    """Decode a SegWit address."""

    if isinstance(address, str):
        address = address.strip()

    # the following check was originally in b32decode
    # but it does not pertain there
    if len(address) > 90:
        raise ValueError(f"Bech32 address length ({len(address)}) > 90")

    hrp, data = b32decode(address)

    # check that it is a SegWit address
    i = _P2W_PREFIXES.index(hrp)

    if len(data) == 0:
        raise ValueError(f"Bech32 address with empty data")

    witvers = data[0]
    witprog = _convertbits(data[1:], 5, 8, False)
    _check_witness(witvers, witprog)

    return _NETWORKS[i], witvers, witprog


def _encode(network: str, wv: int, wp: WitnessProgram) -> bytes:
    """Encode a SegWit address."""

    _check_witness(wv, wp)
    hrp = _P2W_PREFIXES[_NETWORKS.index(network)]
    ret = b32encode(hrp, [wv] + _convertbits(wp, 8, 5))
    return ret


def _p2wpkh_address(h160: bytes, native: bool, network: str) -> bytes:
    """Return the p2wpkh address as native SegWit or legacy p2sh-wrapped."""

    if len(h160) != 20:
        msg = f"Witness program length ({len(h160)}) "
        msg += "is not 20-bytes"
        raise ValueError(msg)
    witvers = 0
    if native:
        return _encode(network, witvers, h160)
    # script_pubkey = [0, key_hash], i.e. 0x0014{20-byte key-hash}
    script_pubkey = b'\x00\x14' + h160
    return p2sh_address(script_pubkey, network)


def _h160_pubkey(pubkey: Octets) -> bytes:
    pubkey = bytes_from_hexstring(pubkey)
    if pubkey[0] not in (2, 3):
        raise ValueError(f"Uncompressed pubkey {pubkey.hex()}")
    psize = 32  # FIXME: parametrize on network
    if len(pubkey) != psize + 1:
        msg = f"Wrong pubkey size: {len(pubkey)} instead of {psize + 1}"
        raise ValueError(msg)
    return hash160(pubkey)


def p2wpkh_address(pubkey: Octets, network: str = 'mainnet') -> bytes:
    """Return the p2wpkh native SegWit address."""
    native = True
    return _p2wpkh_address(_h160_pubkey(pubkey), native, network)


def p2wpkh_p2sh_address(pubkey: Octets, network: str = 'mainnet') -> bytes:
    """Return the p2wpkh-p2sh (legacy) address."""
    native = False
    return _p2wpkh_address(_h160_pubkey(pubkey), native, network)


def _p2wsh_address(h256: bytes, native: bool, network: str) -> bytes:
    """Return the address as native SegWit bech32 or legacy p2sh-wrapped."""

    if len(h256) != 32:
        raise ValueError(f"witness program length ({len(h256)}) is not 32")
    witvers = 0
    if native:
        return _encode(network, witvers, h256)

    script_pubkey: List[Token] = [witvers, h256]
    return p2sh_address(encode(script_pubkey), network)


def p2wsh_address(wscript: Octets, network: str = 'mainnet') -> bytes:
    """Return the p2wsh native SegWit address."""
    native = True
    return _p2wsh_address(sha256(wscript), native, network)


def p2wsh_p2sh_address(wscript: Octets, network: str = 'mainnet') -> bytes:
    """Return the p2wsh-p2sh (legacy) address."""
    native = False
    return _p2wsh_address(sha256(wscript), native, network)


def hash_from_bech32_address(address: Union[str, bytes]) -> Tuple[str, bool, bytes]:

    network, wv, wp = _decode(address)
    if wv != 0:
        raise ValueError(f"Invalid witness version: {wv}")
    if len(wp) == 20:
        is_script_hash = False
    else:
        is_script_hash = True

    return network, is_script_hash, bytes(wp)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import random
from hashlib import sha256 as hf  # FIXME: any hf
from typing import Sequence, Tuple, List, Dict
from collections import defaultdict

from .curve import Point
from .curvemult import mult, double_mult
from .curves import secp256k1 as ec  # FIXME: any curve
from .utils import int_from_bits, octets_from_point, point_from_octets

# FIXME: should be urandom, but then tests would be non-deterministic
random.seed(42)

PubkeyRing = Dict[int, Sequence[Point]]
SValues = Dict[int, Sequence[int]]


def _hash(msg: bytes, R: bytes, i: int, j: int) -> bytes:
    temp = msg + R
    temp += i.to_bytes(4, byteorder='big') + j.to_bytes(4, byteorder='big')
    return hf(temp).digest()


def _get_msg_format(msg: bytes, pubk_rings: PubkeyRing) -> bytes:
    m = msg
    rings = len(pubk_rings)
    for i in range(rings):
        for P in pubk_rings[i]:
            Pbytes = octets_from_point(P, True, ec)
            m += Pbytes
    return hf(m).digest()


def sign(msg: bytes,
         k: Sequence[int],
         sign_key_idx: Sequence[int],
         sign_keys: Sequence[int],
         pubk_rings: PubkeyRing) -> Tuple[bytes, SValues]:
    """Borromean ring signature - signing algorithm

        https://github.com/ElementsProject/borromean-signatures-writeup
        https://github.com/Blockstream/borromean_paper/blob/master/borromean_draft_0.01_9ade1e49.pdf

        inputs:
        - msg: message to be signed (bytes)
        - sign_key_idx: list of indexes representing each signing key per ring
        - sign_keys: list containing the whole set of signing keys (one per ring)
        - pubk_rings: dictionary of sequences representing single rings of pubkeys
    """

    s: Dict[int, Sequence[int]] = defaultdict(list)
    e: Dict[int, Sequence[int]] = defaultdict(list)
    m = _get_msg_format(msg, pubk_rings)
    e0bytes = m
    ring_size = len(pubk_rings)
    # step 1
    for i in range(ring_size):
        keys_size = len(pubk_rings[i])
        s[i] = [0]*keys_size
        e[i] = [0]*keys_size
        j_star = sign_key_idx[i]
        start_idx = (j_star + 1) % keys_size
        R = octets_from_point(mult(k[i]), True, ec)
        if start_idx != 0:
            for j in range(start_idx, keys_size):
                s[i][j] = random.getrandbits(256)
                e[i][j] = int_from_bits(_hash(m, R, i, j), ec)
                assert 0 < e[i][j] < ec.n, "sign fail: how did you do that?!?"
                T = double_mult(-e[i][j], pubk_rings[i][j], s[i][j])
                R = octets_from_point(T, True, ec)
        e0bytes += R
    e0 = hf(e0bytes).digest()
    # step 2
    for i in range(ring_size):
        e[i][0] = int_from_bits(_hash(m, e0, i, 0), ec)
        assert 0 < e[i][0] < ec.n, "sign fail: how did you do that?!?"
        j_star = sign_key_idx[i]
        for j in range(1, j_star+1):
            s[i][j-1] = random.getrandbits(256)
            T = double_mult(-e[i][j-1], pubk_rings[i][j-1], s[i][j-1])
            R = octets_from_point(T, True, ec)
            e[i][j] = int_from_bits(_hash(m, R, i, j), ec)
            assert 0 < e[i][j] < ec.n, "sign fail: how did you do that?!?"
        s[i][j_star] = k[i] + sign_keys[i]*e[i][j_star]
    return e0, s


def verify(msg: bytes, e0: bytes, s: SValues, pubk_rings: PubkeyRing) -> bool:
    """Borromean ring signature - verification algorithm

    inputs: 
    - msg: message to be signed (bytes)
    - e0: pinned e-value needed to start the verification algorithm
    - s: s-values, both real (one per ring) and forged
    - pubk_rings: dictionary of sequences representing single rings of pubkeys
    """

    # this is just a try/except wrapper for the Errors
    # raised by _verify
    try:
        return _verify(msg, e0, s, pubk_rings)
    except Exception:
        return False


def _verify(msg: bytes, e0: bytes, s: SValues, pubk_rings: PubkeyRing) -> bool:

    ring_size = len(pubk_rings)
    m = _get_msg_format(msg, pubk_rings)
    e: Dict[int, Sequence[int]] = defaultdict(list)
    e0bytes = m
    for i in range(ring_size):
        keys_size = len(pubk_rings[i])
        e[i] = [0]*keys_size
        e[i][0] = int_from_bits(_hash(m, e0, i, 0), ec)
        assert e[i][0] != 0, "invalid sig: how did you do that?!?"
        R = b'\0x00'
        for j in range(keys_size):
            T = double_mult(-e[i][j], pubk_rings[i][j], s[i][j])
            R = octets_from_point(T, True, ec)
            if j != len(pubk_rings[i])-1:
                e[i][j+1] = int_from_bits(_hash(m, R, i, j+1), ec)
                assert e[i][j+1] != 0, "invalid sig: how did you do that?!?"
            else:
                e0bytes += R
    e0_prime = hf(e0bytes).digest()
    return e0_prime == e0



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Entropy conversions from/to binary 0/1 string, bytes-like, and int.

Input entropy can be expressed as
binary 0/1 string, bytes-like, or integer.

Output entropy is always a binary 0/1 string.
"""

import math
import secrets
from hashlib import sha256
from typing import Iterable, List, Optional, Union

BinStr = str  # binary 0/1 string
Entropy = Union[BinStr, int, bytes]

_bits = 128, 160, 192, 224, 256


def binstr_from_entropy(entr: Entropy,
                        bits: Union[int, Iterable[int]] = _bits) -> BinStr:
    """Convert the input entropy to binary 0/1 string.

    Input entropy can be expressed as
    binary 0/1 string, bytes-like, or integer;
    by default, it must be 128, 160, 192, 224, or 256 bits.

    In the case of binary 0/1 string and bytes-like
    leading zeros are not considered redundant padding.
    In the case of integer, where leading zeros cannot be represented,
    if the bit length is not an allowed value, then the binary 0/1
    string is padded with leading zeros up to the first allowed bit
    length; if the integer bit length is longer than the maximum
    length, then only the leftmost bits are retained.
    """

    if isinstance(bits, int):
        bits = (bits, )       # if a single int, make it a tuple
    bits = sorted(set(bits))  # ascending unique sorting of allowed bits

    if isinstance(entr, str):
        binstr_entr = entr.strip()
        if binstr_entr[:2] == '0b':
            binstr_entr = binstr_entr[2:]
        int(binstr_entr, 2)  # check that entr is a valid binary string
        nbits = len(binstr_entr)
        # no length adjustment
    elif isinstance(entr, bytes):
        nbits = len(entr) * 8
        int_entr = int.from_bytes(entr, 'big')
        binstr_entr = bin(int_entr)[2:]  # remove '0b'
        # no length adjustment
    elif isinstance(entr, int):
        if entr < 0:
            raise ValueError(f"negative entropy ({entr})")
        binstr_entr = bin(entr)[2:]  # remove '0b'
        nbits = len(binstr_entr)
        if nbits > bits[-1]:
            # only the leftmost bits are retained
            binstr_entr = binstr_entr[:bits[-1]]
            nbits = bits[-1]
        elif nbits not in bits:
            # next allowed bit length
            nbits = next(v for i, v in enumerate(bits) if v > nbits)
    else:
        m = "entropy must be binary 0/1 string, bytes-like, or int; "
        m += f"not '{type(entr).__name__}'"
        raise TypeError(m)

    if nbits not in bits:
        raise ValueError(f"{nbits} bits entropy provided; expected: {bits}")
    return binstr_entr.zfill(nbits)  # might need padding with leading zeros


def generate_entropy(bits: int, dice_base: int = 0,
                     rolls: Optional[List[int]] = None, shuffle: bool = True,
                     hash: bool = True, xor: bool = True) -> BinStr:
    """Return CSPRNG system entropy mixed with exogenous roll-based entropy.

    If no exogenous entropy is provided, then entropy generated with the
    system cryptographically strong pseudo-random number generator (CSPRNG)
    is returned.

    Instead, if exogenous entropy is provided, then it is possibly manipulated
    and, finally, XOR-ed with the CSPRNG system entropy.

    The optional exogenous roll-based entropy must consist of integers in the
    [1-dice_base] range; anyway, only rolls having value in the [1-base] range
    are used, with base being the highest power of 2 lower than the dice_base
    (e.g. for a traditional D6 dice, only rolls having value in [1-4] are
    used; for a D20 dice, only rolls having value in [1-16] are used; etc.).

    If provided, the exogenous roll-based entropy must supply at least the
    required number of bits. Its rolls can be shuffled, entropy can be hashed,
    and it is finally XOR-ed with the CSPRNG system entropy.
    If not shuffled, hashed, and/or XOR-ed, then the function returns the
    rightmost required number of bits from the unaltered exogenous entropy.
    """

    if bits not in _bits:
        raise ValueError(f"Number of bits ({bits}) must be in {_bits}")

    i = 0
    # start with the exogenously provided roll-based entropy
    if rolls is not None:
        if dice_base < 2:
            raise ValueError(f"Invalid dice base ({dice_base}): must be >= 2")
        bits_per_roll = math.floor(math.log2(dice_base))
        # used base
        base = 2 ** bits_per_roll

        if shuffle:
            secrets.SystemRandom().shuffle(rolls)

        min_roll_number = math.ceil(bits/bits_per_roll)
        for r in rolls:
            # collect only usable rolls in [1-base)]
            if 0 < r and r <= base:
                i *= base
                i += r-1
                min_roll_number -= 1
            # reject invalid rolls not in [1-dice_base)]
            elif r < 1 or r > dice_base:
                msg = f"invalid ({r}) roll, not in [1-{dice_base}]"
                raise ValueError(msg)
        if min_roll_number > 0:
            msg = f"too few usable [1-{base}] rolls, missing {min_roll_number}"
            raise ValueError(msg)

        # hash the (possibly shuffled) exogenous entropy
        if hash:
            if i.bit_length() > 256:
                i >>= i.bit_length() - 256
            h256 = sha256(i.to_bytes(32, byteorder='big')).digest()
            i = int.from_bytes(h256, byteorder='big')

    # XOR the (possibly shuffled and/or hashed) exogenous entropy
    # with CSPRNG system entropy
    if xor:
        i ^= secrets.randbits(bits)

    # convert to binary string
    binstr = bin(i)
    # remove leading '0b'
    binstr = binstr[2:]
    # do not lose leading zeros
    binstr = binstr.zfill(bits)
    # take only the (possibly XOR-ed) rightmost bits
    return binstr[-bits:]



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""ScriptPubKey functions.

"""

from typing import Iterable, List, Union, Tuple

from .script import encode, Token
from .utils import Octets, bytes_from_hexstring
from . import segwitaddress, address


def nulldata_scriptPubKey(data: Octets) -> List[Token]:
    """Return the nulldata scriptPubKey with the provided data."""

    data = bytes_from_hexstring(data)
    if len(data) > 40:
        msg = f"Invalid data lenght ({len(data)} bytes) "
        msg += "for nulldata scriptPubKey"
        raise ValueError(msg)
    return ['OP_RETURN', data.hex()]


def p2pk_scriptPubKey(pubkey: Octets) -> List[Token]:
    """Return the p2pk scriptPubKey of the provided pubkey."""

    pubkey = bytes_from_hexstring(pubkey)
    if len(pubkey) not in (33, 65):
        msg = f"Invalid pubkey lenght ({len(pubkey)} bytes) "
        msg += "for p2pk scriptPubKey"
        raise ValueError(msg)
    return [pubkey.hex(), 'OP_CHECKSIG']


def multisig_scriptPubKey(m: int, pubkeys: Iterable[Octets]) -> List[Token]:
    """Return the m-of-n multi-sig scriptPubKey of the provided pubkeys."""

    if m<1 or m>16:
        raise ValueError(f"Invalid m ({m})")
        # raise ValueError("Impossible m-of-n ({m}-of-{n})")

    scriptPubKey : List[Token] = [m]
    for pubkey in pubkeys:
        pubkey = bytes_from_hexstring(pubkey)
        if len(pubkey) not in (33, 65):
            msg = f"Invalid pubkey lenght ({len(pubkey)} bytes) "
            msg += "for m-of-n multi-sig scriptPubKey"
            raise ValueError(msg)
        scriptPubKey.append(pubkey.hex())

    # FIXME: handle script max length
    # FIXME: enable lexicographic key sorting
    n = len(scriptPubKey)-1
    if n<1 or n>16:
        raise ValueError(f"Invalid n ({n})")
    if m>n:
        raise ValueError(f"Impossible m-of-n ({m}-of-{n})")
    scriptPubKey.append(n)
    scriptPubKey.append('OP_CHECKMULTISIGVERIFY')
    return scriptPubKey


def p2pkh_scriptPubKey(pubkey_h160: Octets) -> List[Token]:
    """Return the p2pkh scriptPubKey of the provided HASH160 pubkey-hash."""

    pubkey_h160 = bytes_from_hexstring(pubkey_h160)
    if len(pubkey_h160) != 20:
        msg = f"Invalid pubkey-hash lenght ({len(pubkey_h160)} bytes) "
        msg += "for p2pkh scriptPubKey"
        raise ValueError(msg)
    return ['OP_DUP', 'OP_HASH160', pubkey_h160.hex(), 'OP_EQUALVERIFY', 'OP_CHECKSIG']


def p2sh_scriptPubKey(script_h160: Octets) -> List[Token]:
    """Return the p2sh scriptPubKey of the provided HASH160 script-hash."""

    script_h160 = bytes_from_hexstring(script_h160)
    if len(script_h160) != 20:
        msg = f"Invalid script-hash lenght ({len(script_h160)} bytes) "
        msg += "for p2sh scriptPubKey"
        raise ValueError(msg)
    return ['OP_HASH160', script_h160.hex(), 'OP_EQUAL']


def p2wpkh_scriptPubKey(pubkey_h160: Octets) -> List[Token]:
    """Return the p2wpkh scriptPubKey of the provided HASH160 pubkey-hash.
    
    For P2WPKH, the witness program must be the HASH160 20-byte pubkey-hash;
    the scriptPubkey is the witness version 0 followed by the canonical push
    of the witness program (program lenght + program),
    that is 0x0014{20-byte key-hash}
    """

    pubkey_h160 = bytes_from_hexstring(pubkey_h160)
    if len(pubkey_h160) != 20:
        msg = f"Invalid witness program lenght ({len(pubkey_h160)} bytes) "
        msg += "for p2wpkh scriptPubKey"
        raise ValueError(msg)
    return [0, pubkey_h160.hex()]


def p2wsh_scriptPubKey(script_h160: Octets) -> List[Token]:
    """Return the p2wsh scriptPubKey of the provided SHA256 script-hash.
    
    For P2WSH, the witness program must be the SHA256 32-byte script-hash;
    the scriptPubkey is the witness version 0 followed by the canonical push
    of the witness program (program lenght + program),
    that is 0x0020{32-byte script-hash}
    """

    script_h160 = bytes_from_hexstring(script_h160)
    if len(script_h160) != 32:
        msg = f"Invalid witness program lenght ({len(script_h160)} bytes) "
        msg += "for p2wsh scriptPubKey"
        raise ValueError(msg)
    return [0, script_h160.hex()]


def address_from_scriptPubKey(scriptPubKey: Iterable[Token],
                              network: str = "mainnet") -> bytes:
    """Return the bech32/base58 address from the input scriptPubKey."""

    s = encode(scriptPubKey)
    len_s = len(s)
    if len_s == 34 and s[:2] == b'\x00\x20':
        return segwitaddress._p2wsh_address(s[2:], True, network)
    elif len_s == 22 and s[:2] == b'\x00\x14':
        return segwitaddress._p2wpkh_address(s[2:], True, network)
    elif len_s == 23 and s[:2] == b'\xa9\x14' and s[-1:] == b'\x87':
        return address._p2sh_address(s[2:len_s-1], network)
    elif len_s == 25 and s[:3] == b'\x76\xa9\x14' and s[-2:] == b'\x88\xac':
        return address._p2pkh_address(s[3:len_s-2], network)
    else:
        raise ValueError("Unknown script")


def scriptPubKey_from_address(addr: Union[str, bytes]) -> Tuple[List[Token], str]:
    """Return (scriptPubKey, network) from the input bech32/base58 address"""

    if segwitaddress.has_segwit_prefix(addr):
        # also check witness validity
        network, witvers, witprog = segwitaddress._decode(addr)
        if witvers == 0:
            len_wprog = len(witprog)
            if len_wprog == 32:
                return p2wsh_scriptPubKey(bytes(witprog)), network
            else:  # must be len_wprog == 20
                return p2wpkh_scriptPubKey(bytes(witprog)), network
        else:
            raise ValueError(f"Unhandled witness version ({witvers})")
    else:
        network, is_p2sh, payload = address.h160_from_base58_address(addr)
        if is_p2sh:
            return p2sh_scriptPubKey(payload), network
        else:
            return p2pkh_scriptPubKey(payload), network



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Electrum entropy / mnemonic / seed functions.

Electrum mnemonic is versioned, conveying BIP32 derivation rule too.
"""

import hmac
from hashlib import pbkdf2_hmac, sha512
from typing import Tuple

from . import bip32
from .entropy import BinStr, Entropy, binstr_from_entropy
from .mnemonic import (Mnemonic, _entropy_from_indexes, _indexes_from_entropy,
                       _indexes_from_mnemonic, _mnemonic_from_indexes)

_MNEMONIC_VERSIONS = {
    'standard':  '01',  # P2PKH and Multisig P2SH wallets
    'segwit': '100',  # P2WPKH and P2WSH wallets
    '2fa': '101',  # Two-factor authenticated wallets
    '2fa_segwit': '102',  # Two-factor authenticated wallets, using segwit
}


def version_from_mnemonic(mnemonic: Mnemonic) -> str:
    """Return the Electrum version embedded in the mnemonic sentence."""

    s = hmac.new(b"Seed version", mnemonic.encode('utf8'), sha512).hexdigest()

    if s.startswith(_MNEMONIC_VERSIONS['standard']):
        return 'standard'
    if s.startswith(_MNEMONIC_VERSIONS['segwit']):
        return 'segwit'
    if s.startswith(_MNEMONIC_VERSIONS['2fa']):
        return '2fa'
    if s.startswith(_MNEMONIC_VERSIONS['2fa_segwit']):
        return '2fa_segwit'

    raise ValueError(f"unknown electrum mnemonic version ({s[:3]})")


def mnemonic_from_entropy(electrum_version: str, entropy: Entropy,
                          lang: str = "en") -> Mnemonic:
    """Convert input entropy to versioned Electrum mnemonic sentence.

    Input entropy can be expressed as
    binary 0/1 string, bytes-like, or integer.

    In the case of binary 0/1 string and bytes-like,
    leading zeros are considered redundant padding.
    """

    if electrum_version not in _MNEMONIC_VERSIONS:
        m = f"mnemonic version '{electrum_version}' not in electrum allowed "
        m += f"mnemonic versions {list(_MNEMONIC_VERSIONS.keys())}"
        raise ValueError(m)
    version = _MNEMONIC_VERSIONS[electrum_version]

    binstr_entropy = binstr_from_entropy(entropy)
    int_entropy = int(binstr_entropy, 2)
    invalid = True
    while invalid:
        # electrum considers entropy as integer, losing any leading zero
        # so the value of binstr_entropy before the while must be updated
        nbits = int_entropy.bit_length()
        binstr_entropy = binstr_from_entropy(int_entropy, nbits)
        indexes = _indexes_from_entropy(binstr_entropy, lang)
        mnemonic = _mnemonic_from_indexes(indexes, lang)
        # version validity check
        s = hmac.new(b"Seed version",
                     mnemonic.encode('utf8'), sha512).hexdigest()
        if s.startswith(version):
            invalid = False
        # next trial
        int_entropy += 1

    return mnemonic


def entropy_from_mnemonic(mnemonic: Mnemonic, lang: str = "en") -> BinStr:
    """Convert mnemonic sentence to Electrum versioned entropy."""

    # verify that it is a valid Electrum mnemonic sentence
    _ = version_from_mnemonic(mnemonic)
    indexes = _indexes_from_mnemonic(mnemonic, lang)
    entropy = _entropy_from_indexes(indexes, lang)
    return entropy


def _seed_from_mnemonic(mnemonic: Mnemonic, passphrase: str) -> Tuple[str, bytes]:
    """Return (version, seed) from mnemonic according to Electrum standard."""

    version = version_from_mnemonic(mnemonic)

    hf_name = 'sha512'
    password = mnemonic.encode()
    salt = ('electrum' + passphrase).encode()
    iterations = 2048
    dksize = 64
    return version, pbkdf2_hmac(hf_name, password, salt, iterations, dksize)


def masterxprv_from_mnemonic(mnemonic: Mnemonic,
                             passphrase: str,
                             network: str = 'mainnet') -> bytes:
    """Return BIP32 master extended private key from Electrum mnemonic.

    Note that for a 'standard' mnemonic the derivation path is "m",
    for a 'segwit' mnemonic it is "m/0h" instead.
    """

    version, seed = _seed_from_mnemonic(mnemonic, passphrase)
    prefix = bip32._NETWORKS.index(network)

    if version == 'standard':
        xversion = bip32._XPRV_PREFIXES[prefix]
        return bip32.rootxprv_from_seed(seed, xversion)
    elif version == 'segwit':
        xversion = bip32._P2WPKH_PRV_PREFIXES[prefix]
        rootxprv = bip32.rootxprv_from_seed(seed, xversion)
        return bip32.ckd(rootxprv, 0x80000000)  # "m/0h"
    else:
        raise ValueError(f"Unmanaged electrum mnemonic version ({version})")



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""BIP32 Hierarchical Deterministic Wallet functions.

A deterministic wallet is a hash-chain of private/public key pairs that
derives from a single root, which is the only element requiring backup.
Moreover, there are schemes where public keys can be calculated without
accessing private keys.

A hierarchical deterministic wallet is a tree of multiple hash-chains,
derived from a single root, allowing for selective sharing of keypair
chains.

Here, the HD wallet is implemented according to BIP32 bitcoin standard
https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki.
"""

from hashlib import sha512
from hmac import HMAC
from typing import Dict, List, Optional, Sequence, Tuple, Union

from .address import p2pkh_address
from .base58 import b58decode, b58encode
from .curvemult import mult
from .curves import secp256k1 as ec
from .segwitaddress import p2wpkh_address, p2wpkh_p2sh_address
from .utils import (Octets, bytes_from_hexstring, hash160, int_from_octets,
                    octets_from_point, point_from_octets)
from .wif import wif_from_prvkey

# Bitcoin core uses the m/0h (core) BIP32 derivation path
# with xprv/xpub and tprv/tpub Base58 encoding

# VERSION BYTES (4 bytes)

# m/44h/0h  p2pkh or p2sh
MAIN_xprv = b'\x04\x88\xAD\xE4'
MAIN_xpub = b'\x04\x88\xB2\x1E'
# m/44h/1h  p2pkh or p2sh
TEST_tprv = b'\x04\x35\x83\x94'
TEST_tpub = b'\x04\x35\x87\xCF'

# m/49h/0h  p2wpkh-p2sh (p2sh-wrapped-segwit)
MAIN_yprv = b'\x04\x9D\x78\x78'
MAIN_ypub = b'\x04\x9D\x7C\xB2'
# m/49h/1h  p2wpkh-p2sh (p2sh-wrapped-segwit)
TEST_uprv = b'\x04\x4A\x4E\x28'
TEST_upub = b'\x04\x4A\x52\x62'

#   ---     p2wsh-p2sh (p2sh-wrapped-segwit)
MAIN_Yprv = b'\x02\x95\xB0\x05'
MAIN_Ypub = b'\x02\x95\xB4\x3F'
TEST_Uprv = b'\x02\x42\x85\xB5'
TEST_Upub = b'\x02\x42\x89\xEF'

# m/84h/0h  p2wpkh (native-segwit)
MAIN_zprv = b'\x04\xB2\x43\x0C'
MAIN_zpub = b'\x04\xB2\x47\x46'
# m/84h/1h  p2wpkh (native-segwit)
TEST_vprv = b'\x04\x5F\x18\xBC'
TEST_vpub = b'\x04\x5F\x1C\xF6'

#   ---     p2wsh (native-segwit)
MAIN_Zprv = b'\x02\xAA\x7A\x99'
MAIN_Zpub = b'\x02\xAA\x7E\xD3'
TEST_Vprv = b'\x02\x57\x50\x48'
TEST_Vpub = b'\x02\x57\x54\x83'

_REPEATED_NETWORKS = [
    'mainnet', 'mainnet', 'mainnet', 'mainnet', 'mainnet',
    'testnet', 'testnet', 'testnet', 'testnet', 'testnet',
    'regtest', 'regtest', 'regtest', 'regtest', 'regtest']
_PRV_VERSIONS = [
    MAIN_xprv, MAIN_yprv, MAIN_zprv, MAIN_Yprv, MAIN_Zprv,
    TEST_tprv, TEST_uprv, TEST_vprv, TEST_Uprv, TEST_Vprv,
    TEST_tprv, TEST_uprv, TEST_vprv, TEST_Uprv, TEST_Vprv]
_PUB_VERSIONS = [
    MAIN_xpub, MAIN_ypub, MAIN_zpub, MAIN_Ypub, MAIN_Zpub,
    TEST_tpub, TEST_upub, TEST_vpub, TEST_Upub, TEST_Vpub,
    TEST_tpub, TEST_upub, TEST_vpub, TEST_Upub, TEST_Vpub]

_NETWORKS = ['mainnet', 'testnet', 'regtest']
# p2pkh or p2sh
_XPRV_PREFIXES = [MAIN_xprv, TEST_tprv, TEST_tprv]
_XPUB_PREFIXES = [MAIN_xpub, TEST_tpub, TEST_tpub]
# p2wpkh p2sh-wrapped-segwit
_P2WPKH_P2SH_PRV_PREFIXES = [MAIN_yprv, TEST_uprv, TEST_uprv]
_P2WPKH_P2SH_PUB_PREFIXES = [MAIN_ypub, TEST_upub, TEST_upub]
# p2wsh p2sh-wrapped-segwit
_P2WSH_P2SH_PRV_PREFIXES = [MAIN_Yprv, TEST_Uprv, TEST_Uprv]
_P2WSH_P2SH_PUB_PREFIXES = [MAIN_Ypub, TEST_Upub, TEST_Upub]
# p2wpkh native-segwit
_P2WPKH_PRV_PREFIXES = [MAIN_zprv, TEST_vprv, TEST_vprv]
_P2WPKH_PUB_PREFIXES = [MAIN_zpub, TEST_vpub, TEST_vpub]
# p2wsh native-segwit
_P2WSH_PRV_PREFIXES = [MAIN_Zprv, TEST_Vprv, TEST_Vprv]
_P2WSH_PUB_PREFIXES = [MAIN_Zpub, TEST_Vpub, TEST_Vpub]


# [  : 4] version
# [ 4: 5] depth
# [ 5: 9] parent_fingerprint
# [ 9:13] index
# [13:45] chain_code
# [45:78] key (private/public)

def _check_version_key(v: bytes, k: bytes) -> None:
    if not isinstance(k, bytes):
        msg = f"invalid key type: must be bytes, not '{type(k).__name__}'"
        raise TypeError(msg)
    if len(k) != 33:
        raise ValueError(f"invalid {len(k)}-bytes key length")

    if v in _PRV_VERSIONS:
        if k[0] != 0:
            raise ValueError("prvversion/pubkey mismatch")
    elif v in _PUB_VERSIONS:
        if k[0] not in (2, 3):
            raise ValueError("pubversion/prvkey mismatch")
    else:
        raise ValueError("unknown extended key version")


def _check_depth_fingerprint_index(d: int, f: bytes, i: bytes) -> None:
    if not isinstance(f, bytes):
        msg = "invalid parent_fingerprint type: "
        msg += f"must be bytes, not '{type(f).__name__}'"
        raise TypeError(msg)
    if len(f) != 4:
        raise ValueError(f"invalid {len(f)}-bytes parent_fingerprint length")

    if not isinstance(i, bytes):
        msg = "invalid index type: "
        msg += f"must be bytes, not '{type(i).__name__}'"
        raise TypeError(msg)
    if len(i) != 4:
        raise ValueError(f"invalid {len(i)}-bytes index length")

    if d < 0 or d > 255:
        raise ValueError(f"invalid depth ({d})")
    elif d == 0:
        if f != b'\x00\x00\x00\x00':
            msg = f"zero depth with non-zero parent_fingerprint {f!r}"
            raise ValueError(msg)
        if i != b'\x00\x00\x00\x00':
            msg = f"zero depth with non-zero index {i!r}"
            raise ValueError(msg)
    else:
        if f == b'\x00\x00\x00\x00':
            msg = f"non-zero depth ({d}) with zero parent_fingerprint {f!r}"
            raise ValueError(msg)


def parse(xkey: Octets) -> Dict:

    xkey = b58decode(xkey, 78)
    d = {
        'version'            : xkey[:4],
        'depth'              : xkey[4],
        'parent_fingerprint' : xkey[5:9],
        'index'              : xkey[9:13],
        'chain_code'         : xkey[13:45],
        'key'                : xkey[45:]
    }
    # coherence checks
    _check_version_key(d['version'], d['key'])
    _check_depth_fingerprint_index(d['depth'],
                                   d['parent_fingerprint'], d['index'])
    return d


def serialize(d: Dict) -> bytes:

    # coherence checks
    _check_version_key(d['version'], d['key'])
    _check_depth_fingerprint_index(d['depth'],
                                   d['parent_fingerprint'], d['index'])
    t = d['version']
    t += d['depth'].to_bytes(1, 'big')
    t += d['parent_fingerprint']
    t += d['index']

    chain_code = d['chain_code']
    if not isinstance(chain_code, bytes):
        msg = "invalid chain_code type: "
        msg += f"must be bytes, not '{type(chain_code).__name__}'"
        raise TypeError(msg)
    if len(chain_code) != 32:
        raise ValueError(f"invalid {len(chain_code)}-bytes chain_code length")
    t += chain_code

    t += d['key']

    return b58encode(t)


def parent_fingerprint(xkey: Octets) -> bytes:
    d = parse(xkey)
    if d['depth'] == 0:
        raise ValueError("master key provided")
    return d['parent_fingerprint']


def index(xkey: Octets) -> bytes:
    d = parse(xkey)
    if d['depth'] == 0:
        raise ValueError("master key provided")
    return d['index']


def fingerprint(xkey: Octets) -> bytes:
    d = parse(xkey)
    if d['key'][0] == 0:
        P = mult(int_from_octets(d['key'][1:]))
        d['key'] = octets_from_point(P, True, ec)
    return hash160(d['key'])[:4]


def address_from_xpub(xpub: Octets) -> bytes:
    """Return the address according to the xpub SLIP32 version type."""

    d = parse(xpub)

    if d['key'][0] not in (2, 3):
        raise ValueError("xkey is not a public one")

    if d['version'] in _XPUB_PREFIXES:
        # p2pkh
        return _p2pkh_address_from_xpub(d['version'], d['key'])
    elif d['version'] in _P2WPKH_PUB_PREFIXES:
        # p2wpkh native-segwit
        return _p2wpkh_address_from_xpub(d['version'], d['key'])
    else:
        # v in _P2WPKH_P2SH_PUB_PREFIXES
        # p2wpkh p2sh-wrapped-segwit
        return _p2wpkh_p2sh_address_from_xpub(d['version'], d['key'])


def wif_from_xprv(xprv: Octets, compressed: bool = True) -> bytes:
    """Return the WIF according to xpub version type."""

    d = parse(xprv)

    if d['key'][0] != 0:
        raise ValueError("xkey is not a private one")

    network = _REPEATED_NETWORKS[_PRV_VERSIONS.index(d['version'])]
    return wif_from_prvkey(d['key'][1:], compressed, network)


def _p2pkh_address_from_xpub(v: bytes, pk: bytes) -> bytes:
    network = _REPEATED_NETWORKS[_PUB_VERSIONS.index(v)]
    return p2pkh_address(pk, network)


def p2pkh_address_from_xpub(xpub: Octets) -> bytes:
    """Return the p2pkh address."""
    d = parse(xpub)
    if d['key'][0] not in (2, 3):
        # Deriving pubkey from prvkey would not be enough:
        # compressed ot uncompressed?
        raise ValueError("xkey is not a public one")
    return _p2pkh_address_from_xpub(d['version'], d['key'])


def _p2wpkh_address_from_xpub(v: bytes, pk: bytes) -> bytes:
    network = _REPEATED_NETWORKS[_PUB_VERSIONS.index(v)]
    return p2wpkh_address(pk, network)


def p2wpkh_address_from_xpub(xpub: Octets) -> bytes:
    """Return the p2wpkh (native SegWit) address."""
    d = parse(xpub)
    if d['key'][0] not in (2, 3):
        # if pubkey would be derived from prvkey,
        # then this safety check might be removed
        raise ValueError("xkey is not a public one")
    return _p2wpkh_address_from_xpub(d['version'], d['key'])


def _p2wpkh_p2sh_address_from_xpub(v: bytes, pk: bytes) -> bytes:
    network = _REPEATED_NETWORKS[_PUB_VERSIONS.index(v)]
    return p2wpkh_p2sh_address(pk, network)


def p2wpkh_p2sh_address_from_xpub(xpub: Octets) -> bytes:
    """Return the p2wpkh p2sh-wrapped (legacy) address."""
    d = parse(xpub)
    if d['key'][0] not in (2, 3):
        # if pubkey would be derived from prvkey,
        # then this safety check might be removed
        raise ValueError("xkey is not a public one")
    return _p2wpkh_p2sh_address_from_xpub(d['version'], d['key'])


def rootxprv_from_seed(seed: Octets, version: Octets = MAIN_xprv) -> bytes:
    """derive the BIP32 root master extended private key from the seed"""

    seed = bytes_from_hexstring(seed)
    hd = HMAC(b"Bitcoin seed", seed, sha512).digest()
    d = {
        'version'            : bytes_from_hexstring(version),
        'depth'              : 0,
        'parent_fingerprint' : b'\x00\x00\x00\x00',
        'index'              : b'\x00\x00\x00\x00',
        'chain_code'         : hd[32:],
        'key'                : b'\x00' + hd[:32]
    }
    return serialize(d)


def xpub_from_xprv(xprv: Octets) -> bytes:
    """Neutered Derivation (ND).

    Derivation of the extended public key corresponding to an extended
    private key (“neutered” as it removes the ability to sign transactions).
    """

    d = parse(xprv)

    if d['key'][0] != 0:
        raise ValueError("extended key is not a private one")
    P = mult(int_from_octets(d['key'][1:]))
    d['key'] = octets_from_point(P, True, ec)

    d['version'] = _PUB_VERSIONS[_PRV_VERSIONS.index(d['version'])]

    return serialize(d)


def ckd(xparentkey: Octets, index: Union[Octets, int]) -> bytes:
    """Child Key Derivation (CDK)

    Key derivation is normal if the extended parent key is public or
    index is less than 0x80000000.

    Key derivation is hardened if the extended parent key is private and
    index is not less than 0x80000000.
    """

    d = parse(xparentkey)
    _ckd(d, index)
    return serialize(d)


def _ckd(d: Dict, index: Union[Octets, int]) -> None:
    if isinstance(index, int):
        index = index.to_bytes(4, byteorder='big')

    index = bytes_from_hexstring(index)

    if len(index) != 4:
        raise ValueError(f"a 4 bytes int is required, not {len(index)}")

    d['depth'] += 1

    if d['key'][0] == 0:                             # parent is a prvkey
        parent = int.from_bytes(d['key'][1:], byteorder='big')
        Parent = mult(parent)
        Parent_bytes = octets_from_point(Parent, True, ec)
    else:                                            # parent is a pubkey
        Parent = point_from_octets(d['key'], ec)
        Parent_bytes = d['key']
        if index[0] >= 0x80:                         # hardened derivation
            raise ValueError("hardened derivation from pubkey is impossible")
    d['parent_fingerprint'] = hash160(Parent_bytes)[:4]
    d['index'] = index

    if d['key'][0] == 0:                             # parent is a prvkey
        if index[0] >= 0x80:                         # hardened derivation
            h = HMAC(d['chain_code'], d['key'] + index, sha512).digest()
        else:                                        # normal derivation
            h = HMAC(d['chain_code'], Parent_bytes + index, sha512).digest()
        d['chain_code'] = h[32:]
        offset = int.from_bytes(h[:32], byteorder='big')
        child = (parent + offset) % ec.n
        d['key'] = b'\x00' + child.to_bytes(32, 'big')
    else:                                            # parent is a pubkey
        h = HMAC(d['chain_code'], d['key'] + index, sha512).digest()
        d['chain_code'] = h[32:]
        offset = int.from_bytes(h[:32], byteorder='big')
        Offset = mult(offset)
        Child = ec.add(Parent, Offset)
        d['key'] = octets_from_point(Child, True, ec)


def _indexes_from_path(path: str) -> Tuple[Sequence[int], bool]:
    """Extract derivation indexes from a derivation path.

    Derivation path must be like "m/44'/0'/1'/0/10" (absolute)
    or "./0/10" (relative).
    """

    steps = path.split('/')
    if steps[0] in ('m', 'M'):
        absolute = True
    elif steps[0] == '.':
        absolute = False
    elif steps[0] == '':
        raise ValueError('Empty derivation path')
    else:
        raise ValueError(f'Invalid derivation path root: "{steps[0]}"')
    if len(steps) > 256:
        raise ValueError(f'Derivation path depth {len(steps)-1}>255')

    indexes: List[int] = list()
    for step in steps[1:]:
        hardened = False
        if step[-1] in ("'", "H", "h"):
            hardened = True
            step = step[:-1]
        index = int(step)
        index += 0x80000000 if hardened else 0
        indexes.append(index)

    return indexes, absolute


def derive(xkey: Octets, path: Union[str, Sequence[int]]) -> bytes:
    """Derive an extended key.

    Derivation is according to absolute path like "m/44h/0'/1H/0/10"
    or relative path like "./0/10".
    """

    d = parse(xkey)

    if isinstance(path, str):
        path = path.strip()
        indexes, absolute = _indexes_from_path(path)
        if absolute and d["depth"] != 0:
            msg = "Absolute derivation path for non-root master key"
            raise ValueError(msg)
    else:
        indexes = path

    final_depth = d["depth"] + len(indexes)
    if final_depth > 255:
        raise ValueError(f'Derivation path final depth {final_depth}>255')

    for index in indexes:
        _ckd(d, index)

    return serialize(d)


def crack(parent_xpub: Octets, child_xprv: Octets) -> bytes:
    p = parse(parent_xpub)

    if p['key'][0] not in (2, 3):
        raise ValueError("extended parent key is not a public one")

    c = parse(child_xprv)
    if c['key'][0] != 0:
        raise ValueError("extended child key is not a private one")

    # check depth
    if c['depth'] != p['depth'] + 1:
        raise ValueError("not a parent's child: wrong depth relation")

    # check fingerprint
    if c['parent_fingerprint'] != hash160(p['key'])[:4]:
        raise ValueError("not a parent's child: wrong parent fingerprint")

    # check normal derivation
    if c['index'][0] >= 0x80:
        raise ValueError("hardened child derivation")

    p['version'] = c['version']

    h = HMAC(p['chain_code'], p['key'] + c['index'], sha512).digest()
    offset = int.from_bytes(h[:32], byteorder='big')
    child = int.from_bytes(c['key'][1:], byteorder='big')
    parent = (child - offset) % ec.n
    Parent_b = b'\x00' + parent.to_bytes(32, byteorder='big')
    p['key'] = Parent_b

    return serialize(p)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

from typing import Tuple, Union

from .address import p2pkh_address
from .base58 import b58decode, b58encode
from .curvemult import mult
from .curves import secp256k1
from .segwitaddress import p2wpkh_address, p2wpkh_p2sh_address
from .utils import (Octets, bytes_from_hexstring, int_from_octets,
                    octets_from_int, octets_from_point)

_NETWORKS = ['mainnet', 'testnet', 'regtest']
_CURVES = [secp256k1, secp256k1, secp256k1]
_WIF_PREFIXES = [
    b'\x80',  # WIF starts with {K,L} (if compressed) or 5 (if uncompressed)
    b'\xef',  # WIF starts with c (if compressed) or 9 (if uncompressed)
    b'\xef',  # WIF starts with c (if compressed) or 9 (if uncompressed)
]
_P2PKH_PREFIXES = [
    b'\x00',  # address starts with 1
    b'\x6f',  # address starts with {m, n}
    b'\x6f'   # address starts with {m, n}
]
_P2SH_PREFIXES = [
    b'\x05',  # address starts with 3
    b'\xc4',  # address starts with 2
    b'\xc4',  # address starts with 2
]

def wif_from_prvkey(prv: Union[int, Octets],
                    compressed: bool = True,
                    network: str = 'mainnet') -> bytes:
    """Return the Wallet Import Format from a private key."""

    network_index = _NETWORKS.index(network)
    payload = _WIF_PREFIXES[network_index]

    ec = _CURVES[network_index]
    prv = bytes_from_hexstring(prv)

    if isinstance(prv, bytes):
        if len(prv) != 32:  # TODO: parametrize on ec
            raise ValueError(f"not a 32-bytes private key: {prv.hex()}")
        prv = int.from_bytes(prv, byteorder='big')
    if not 0 < prv < ec.n:
        raise ValueError(f"private key {hex(prv)} not in (0, ec.n)")

    payload += octets_from_int(prv, ec.nsize)
    payload += b'\x01' if compressed else b''
    return b58encode(payload)


def prvkey_from_wif(wif: Union[str, bytes]) -> Tuple[int, bool, str]:
    """Return the (private key, compressed, network) tuple from a WIF."""

    if isinstance(wif, str):
        wif = wif.strip()

    payload = b58decode(wif)
    wif_index = _WIF_PREFIXES.index(payload[0:1])
    ec = _CURVES[wif_index]

    if len(payload) == ec.nsize + 2:       # compressed WIF
        compressed = True
        if payload[-1] != 0x01:            # must have a trailing 0x01
            raise ValueError("Not a compressed WIF: missing trailing 0x01")
        prvkey = payload[1:-1]
        prv = int_from_octets(prvkey)
    elif len(payload) == ec.nsize + 1:     # uncompressed WIF
        compressed = False
        prvkey = payload[1:]
        prv = int_from_octets(prvkey)
    else:
        raise ValueError(f"Not a WIF: wrong size ({len(payload)})")

    if not 0 < prv < ec.n:
        msg = f"Not a WIF: private key {hex(prv)} not in [1, n-1]"
        raise ValueError(msg)

    network = _NETWORKS[wif_index]
    return prv, compressed, network


def _pubkey_from_wif(wif: Union[str, bytes]) -> Tuple[bytes, str]:

    prv, compressed, network = prvkey_from_wif(wif)
    network_index = _NETWORKS.index(network)
    ec = _CURVES[network_index]
    Pub = mult(prv, ec.G, ec)
    o = octets_from_point(Pub, compressed, ec)
    return o, network


def p2pkh_address_from_wif(wif: Union[str, bytes]) -> bytes:
    """Return the address corresponding to a WIF.

    WIF encodes the information about the pubkey to be used for the
    address computation being the compressed or uncompressed one.
    """

    prv, compressed, network = prvkey_from_wif(wif)
    network_index = _NETWORKS.index(network)
    ec = _CURVES[network_index]
    Pub = mult(prv, ec.G, ec)
    o = octets_from_point(Pub, compressed, ec)
    return p2pkh_address(o, network)


def p2wpkh_address_from_wif(wif: Union[str, bytes]) -> bytes:
    o, network = _pubkey_from_wif(wif)
    return p2wpkh_address(o, network)


def p2wpkh_p2sh_address_from_wif(wif: Union[str, bytes]) -> bytes:
    o, network = _pubkey_from_wif(wif)
    return p2wpkh_p2sh_address(o, network)



#!/usr/bin/env python3

# Copyright (C) 2019-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

""" Bitcoin address-based compact signature for messages.

Bitcoin uses an address-based scheme for message signature:
such a signature does prove the control of the private key corresponding to
the address and, consequently, of the associated bitcoins (if any).
Message signature adopts a custom compact 65-bytes serialization format
(i.e. not the ASN.1 DER format used for transactions,
which would results in 71-bytes average signature).

One should never sign a vague statement that could be reused
out of the context it was intended for. Always include at least:

- name (nickname, customer id, e-mail, etc.)
- date and time
- who the message is intended for (name, business name, e-mail, etc.)
- specific purpose of the message

To mitigate the risk of signing a possibly deceiving message,
for any given message a *magic* "Bitcoin Signed Message" prefix is added,
then the hash of the resulting message is ECDSA signed.

As it is the case for all digital signature schemes,
this scheme works with private/public key pairs, not addresses:
the address is only used to uniquely identify a private/public key pair.
At signing time, a wallet infrastructure is required to access the
private key corresponding to a given address;
alternatively, the private key must be provided explicitly.

To verify the signature the public key is not needed
because (EC)DSA allows public key recovery:
public keys that correctly verify the signature
can be implied from the signature itself.
In the case of the Bitcoin secp256k1 curve,
two public keys are recovered
(up to four with non-zero but negligible probability);
at verification time the address must match
that public key in the recovery set
marked as the right one at signature time.

The (r, s) DSA signature is serialized as
[1-byte recovery flag][32-bytes r][32-bytes s],
where the first byte is the recovery flag used
at verification time to discriminate among recovered
public keys (and address types).
Explicitly, the recovery flag value is:

    key_id + (4 if compressed else 0) + 27

where:

- key_id is the index in the [0, 3] range identifying which of the
  recovered public keys is the one associated to the address;
- compressed indicates if the address is the hash of the compressed
  public key representation
- 27 identify a P2PKH address, which is the only kind of address
  supported by Bitcoin Core;
  when the recovery flag is in the [31, 34] range of compressed addresses,
  Electrum also check for P2WPKH-P2SH and P2WPKH
  (SegWit always uses compressed public keys);
  BIP137 (Trezor) uses, instead, 35 and 39 instead of 27
  for P2WPKH-P2SH and P2WPKH (respectively).

+---------+--------+---------------------------------------------------------+
| recflag | key_id | address type                                            |
+=========+========+=========================================================+
|    27   |    0   | P2PKH uncompressed                                      |
+---------+--------+---------------------------------------------------------+
|    28   |    1   | P2PKH uncompressed                                      |
+---------+--------+---------------------------------------------------------+
|    29   |    2   | P2PKH uncompressed                                      |
+---------+--------+---------------------------------------------------------+
|    30   |    3   | P2PKH uncompressed                                      |
+---------+--------+---------------------------------------------------------+
|    31   |    0   | P2PKH compressed (also P2WPKH-P2SH/P2WPKH for Electrum) |
+---------+--------+---------------------------------------------------------+
|    32   |    1   | P2PKH compressed (also P2WPKH-P2SH/P2WPKH for Electrum) |
+---------+--------+---------------------------------------------------------+
|    33   |    2   | P2PKH compressed (also P2WPKH-P2SH/P2WPKH for Electrum) |
+---------+--------+---------------------------------------------------------+
|    34   |    3   | P2PKH compressed (also P2WPKH-P2SH/P2WPKH for Electrum) |
+---------+--------+---------------------------------------------------------+
|    35   |    0   | BIP137 (Trezor) P2WPKH-P2SH                             |
+---------+--------+---------------------------------------------------------+
|    36   |    1   | BIP137 (Trezor) P2WPKH-P2SH                             |
+---------+--------+---------------------------------------------------------+
|    37   |    2   | BIP137 (Trezor) P2WPKH-P2SH                             |
+---------+--------+---------------------------------------------------------+
|    38   |    3   | BIP137 (Trezor) P2WPKH-P2SH                             |
+---------+--------+---------------------------------------------------------+
|    39   |    0   | BIP137 (Trezor) P2WPKH                                  |
+---------+--------+---------------------------------------------------------+
|    40   |    1   | BIP137 (Trezor) P2WPKH                                  |
+---------+--------+---------------------------------------------------------+
|    41   |    2   | BIP137 (Trezor) P2WPKH                                  |
+---------+--------+---------------------------------------------------------+
|    42   |    3   | BIP137 (Trezor) P2WPKH                                  |
+---------+--------+---------------------------------------------------------+

Finally, the signature is base64-encoded to transport it
across channels that are designed to deal with textual data.
Base64-encoding uses 10 digits, 26 lowercase characters, 26 uppercase
characters, '+' (plus sign), and '/' (forward slash).
The equal sign '=' is used as end marker of the encoded message.

https://bitcoin.stackexchange.com/questions/10759/how-does-the-signature-verification-feature-in-bitcoin-qt-work-without-a-public

https://bitcoin.stackexchange.com/questions/12554/why-the-signature-is-always-65-13232-bytes-long

https://bitcoin.stackexchange.com/questions/34135/what-is-the-strmessagemagic-good-for

https://bitcoin.stackexchange.com/questions/36838/why-does-the-standard-bitcoin-message-signature-include-a-message-prefix

https://bitcoin.stackexchange.com/questions/68844/explicit-message-length-in-bitcoin-signed-message

https://github.com/bitcoinjs/bitcoinjs-lib/blob/1079bf95c1095f7fb018f6e4757277d83b7b9d07/src/message.js#L13

https://bitcointalk.org/index.php?topic=6428

https://bitcointalk.org/index.php?topic=6430

https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106?newreg=670c5855241d4340af0cbbc960fd2dc3

https://github.com/bitcoin/bitcoin/pull/524

https://www.reddit.com/r/Bitcoin/comments/bgcgs2/can_bitcoin_core_0171_sign_message_from_segwit/

https://github.com/bitcoin/bips/blob/master/bip-0137.mediawiki

https://github.com/brianddk/bips/blob/legacysignverify/bip-0xyz.mediawiki

"""

from base64 import b64decode, b64encode
from hashlib import sha256
from typing import Optional, Tuple, Union

from .address import h160_from_base58_address, p2pkh_address
from .curvemult import mult
from .dsa import _check_sig, pubkey_recovery, sign
from .segwitaddress import (hash_from_bech32_address, p2wpkh_address,
                            p2wpkh_p2sh_address)
from .utils import hash160, octets_from_point
from .wif import prvkey_from_wif


def _magic_hash(msg: Union[str, bytes]) -> bytes:
    t = b'\x18Bitcoin Signed Message:\n' + len(msg).to_bytes(1, 'big')
    if isinstance(msg, str):
        # Electrum does strip leading and trailing spaces;
        # Bitcoin Core does not
        # msg = msg.strip()
        t += msg.encode()
    else:
        t += msg
    return sha256(t).digest()


def _serialize(rf: int, r: int, s: int) -> bytes:
    # serialize [1-byte rf][32-bytes r][32-bytes s]
    if rf < 27 or rf > 42:
        raise ValueError(f"Invalid recovery flag: {rf}")
    _check_sig((r, s))
    t = bytes([rf]) + r.to_bytes(32, 'big') + s.to_bytes(32, 'big')
    return b64encode(t)


def _deserialize(sig: Union[str, bytes]) -> Tuple[int, int, int]:
    # deserialize [1-byte rf][32-bytes r][32-bytes s]
    sig = b64decode(sig)
    if len(sig) != 65:
        raise ValueError(f"Wrong signature length: {len(sig)} instead of 65")
    rf = sig[0]
    if rf < 27 or rf > 42:
        raise ValueError(f"Invalid recovery flag: {rf}")
    r = int.from_bytes(sig[1:33], byteorder='big')
    s = int.from_bytes(sig[33:], byteorder='big')
    _check_sig((r, s))
    return rf, r, s


def msgsign(msg: Union[str, bytes], wif: Union[str, bytes], 
            addr: Optional[Union[str, bytes]] = None) -> bytes:
    """Generate address-based compact signature for the provided message."""

    if isinstance(addr, str):
        addr = addr.strip()
        addr = addr.encode("ascii")

    # first sign the message
    magic_msg = _magic_hash(msg)
    q, compressed, _ = prvkey_from_wif(wif)
    sig = sign(magic_msg, q)

    # now calculate the key_id
    pubkeys = pubkey_recovery(magic_msg, sig)
    Q = mult(q)
    key_id = pubkeys.index(Q)
    pubkey = octets_from_point(Q, compressed)

    # finally, calculate the recovery flag
    if addr is None or addr == p2pkh_address(pubkey):
        rf = key_id + 27
        rf += 4 if compressed else 0
    # BIP137
    elif addr == p2wpkh_p2sh_address(pubkey):
        rf = key_id + 35
    elif addr == p2wpkh_address(pubkey):
        rf = key_id + 39
    else:
        raise ValueError("Mismatch between private key and address")
    
    return _serialize(rf, sig[0], sig[1]) 


def verify(msg: Union[str, bytes],
           addr: Union[str, bytes], sig: Union[str, bytes]) -> bool:
    """Verify address-based compact signature for the provided message."""

    # try/except wrapper for the Errors raised by _verify
    try:
        return _verify(msg, addr, sig)
    except Exception:
        return False


def _verify(msg: Union[str, bytes],
            addr: Union[str, bytes], sig: Union[str, bytes]) -> bool:
    # Private function for test/dev purposes
    # It raises Errors, while verify should always return True or False

    rf, r, s = _deserialize(sig)
    magic_msg = _magic_hash(msg)
    Qs = pubkey_recovery(magic_msg, (r, s))
    # key_id can be retireved as least significant 2 bits of the recovery flag
    #    key_id = 00;     key_id = 01;     key_id = 10;     key_id = 11
    # 27-27 = 000000;  28-27 = 000001;  29-27 = 000010;  30-27 = 000011
    # 31-27 = 000100;  32-27 = 000101;  33-27 = 000110;  34-27 = 000111
    # 35-27 = 001000;  36-27 = 001001;  37-27 = 001010;  38-27 = 001011
    # 39-27 = 001100;  40-27 = 001101;  41-27 = 001110;  42-27 = 001111
    key_id = rf - 27 & 0b11
    Q = Qs[key_id]

    try:
        _, is_script_hash, h160 = h160_from_base58_address(addr)
        is_b58 = True
    except Exception:
        _, is_script_hash, h160 = hash_from_bech32_address(addr)
        is_b58 = False

    # signature is valid only if the provided address is matched
    compressed = True
    if rf < 31:
        compressed = False
    pubkey = octets_from_point(Q, compressed)
    if is_b58:
        if is_script_hash and 30 < rf and rf < 39:  # P2WPKH-P2SH
            script_pubkey = b'\x00\x14' + hash160(pubkey)
            return hash160(script_pubkey) == h160
        elif rf < 35:                               # P2PKH
            return hash160(pubkey) == h160
        else:
            errmsg = f"Invalid recovery flag ({rf}) for base58 address ({addr!r})"
            raise ValueError(errmsg)
    else:
        if rf > 38 or (30 < rf and rf < 35):        # P2WPKH
            return hash160(pubkey) == h160
        else:
            errmsg = f"Invalid recovery flag ({rf}) for bech32 address ({addr!r})"
            raise ValueError(errmsg)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Elliptic curve class and functions."""

from math import sqrt
from typing import NamedTuple, Tuple

from .numbertheory import mod_inv, mod_sqrt, legendre_symbol

# infinity point in affine coordinates is Inf = (int, 0)
# as no affine point has y=0 coordinate in a group of prime order n
# it can be checked with 'Inf[1] == 0'
Point = Tuple[int, int]

# infinity point in Jacobian coordinates is Inf = (int, int, 0)
# it can be checked with 'Inf[2] == 0'
_JacPoint = Tuple[int, int, int]


def _jac_from_aff(Q: Point) -> _JacPoint:
    # point is assumed to be on curve
    if Q[1] == 0:  # Infinity point in affine coordinates
        return 1, 1, 0
    return Q[0], Q[1], 1


class Curve:
    """Elliptic curve y^2 = x^3 + a*x + b over Fp group."""

    def __init__(self, p: int, a: int, b: int, G: Point, n: int,
                 h: int, t: int, weakness_check: bool = True) -> None:
        # Parameters are checked according to SEC 1 v.2 3.1.1.2.1

        # 1) check that p is an odd prime
        if p % 2 == 0:
            raise ValueError(f"p ({hex(p)}) is not odd")
        # Fermat test will do as _probabilistic_ primality test...
        if not pow(2, p-1, p) == 1:
            raise ValueError(f"p ({hex(p)}) is not prime")

        # 1) check that p has enough bits
        plen = p.bit_length()
        if t != 0:
            t_range = [56, 64, 80, 96, 112, 128, 160, 192, 256]
            if t not in t_range:
                m = f"required security level ({t}) "
                m += f"not in the allowed range {t_range}"
                raise UserWarning(m)
            if plen < t*2:
                m = f"not enough bits ({plen}) for required security level {t}"
                raise UserWarning(m)
        self.t = t

        self.psize = (plen + 7) // 8
        # must be true to break simmetry using quadratic residue
        self.pIsThreeModFour = (p % 4 == 3)
        self._p = p

        # 2. check that a and b are integers in the interval [0, p−1]
        if not 0 <= a < p:
            raise ValueError(f"invalid a ({hex(a)}) for given p ({hex(p)})")
        if not 0 <= b < p:
            raise ValueError(f"invalid b ({hex(b)}) for given p ({hex(p)})")

        # 3. Check that 4*a^3 + 27*b^2 ≠ 0 (mod p).
        d = 4*a*a*a+27*b*b
        if d % p == 0:
            raise ValueError("zero discriminant")
        self._a = a
        self._b = b

        # 2. check that xG and yG are integers in the interval [0, p−1]
        # 4. Check that yG^2 = xG^3 + a*xG + b (mod p).
        if len(G) != 2:
            raise ValueError("Generator must a be a tuple[int, int]")
        if not self.is_on_curve(G):
            raise ValueError("Generator is not on the 'x^3 + a*x + b' curve")
        self.G = int(G[0]), int(G[1])
        self.GJ = self.G[0], self.G[1], 1  # Jacobian coordinates

        # 5. Check that n is prime.
        if n < 2 or (n > 2 and not pow(2, n-1, n) == 1):
            raise ValueError(f"n ({hex(n)}) is not prime")
        delta = int(2 * sqrt(p))
        # also check n with Hasse Theorem
        if h < 2:
            if not (p + 1 - delta <= n <= p + 1 + delta):
                m = f"n ({hex(n)}) not in [p + 1 - delta, p + 1 + delta]"
                raise ValueError(m)
        self.n = n
        self.nlen = n.bit_length()
        self.nsize = (self.nlen + 7) // 8

        # 6. Check cofactor
        exp_h = int(1/n + delta/n + p/n)
        if h != exp_h:
            raise ValueError(f"h ({h}) not as expected ({exp_h})")
        assert t == 0 or h <= pow(2, t/8), f"h ({h}) too big for t ({t})"
        self.h = h

        # 7. Check that nG = Inf.
        # it cannot be simply checked with:
        # Inf = mult(self, n, self.G)
        # as the above would be tautologically true
        InfMinusG = self._mult_aff(n-1, self.G)
        Inf = self.add(InfMinusG, self.G)
        if Inf[1] != 0:
            raise ValueError(f"n ({hex(n)}) is not the group order")

        # 8. Check that n ≠ p
        assert n != p, f"n=p ({hex(n)}) -> weak curve"
        #    raise UserWarning("n=p -> weak curve")
        if weakness_check:
            # 8. Check that p^i % n ≠ 1 for all 1≤i<100
            for i in range(1, 100):
                if pow(p, i, n) == 1:
                    raise UserWarning("weak curve")

    def __str__(self) -> str:
        result = "Curve"
        result += f"\n p   = {hex(self._p)}"
        result += f"\n a   = {hex(self._a)}"
        result += f"\n b   = {hex(self._b)}"
        result += f"\n G.x = {hex(self.G[0])}"
        result += f"\n G.y = {hex(self.G[1])}"
        result += f"\n n   = {hex(self.n)}"
        result += f"\n h = {self.h}"
        result += f"\n t = {self.t}"
        return result

    def __repr__(self) -> str:
        result = "Curve("
        result += f"{hex(self._p)}"
        result += f", {hex(self._a)}, {hex(self._b)}"
        result += f", ({hex(self.G[0])}, {hex(self.G[1])})"
        result += f", {hex(self.n)}"
        result += f", {self.h}"
        result += f", {self.t})"
        return result

    # mult could be a function, but it is used by the Curve constructor;
    # moreover, it might be convenient to provide the Curve class with a
    # basic multiplication method, reserving more advanced one as functions
    def mult(self, m: int, Q: Point = None) -> Point:
        if Q is None:
            Q = self.G
        else:
            self.require_on_curve(Q)
        return self._mult_aff(m, Q)

    def _mult_aff(self, m: int, Q: Point) -> Point:
        # double & add in affine coordinates, using binary decomposition of m
        # Point is assumed to be on curve

        m %= self.n
        if m == 0 or Q[1] == 0:          # Infinity point, affine coordinates
            return 1, 0                  # return Infinity point
        R = 1, 0                         # initialize as infinity point
        while m > 0:                     # use binary representation of m
            if m & 1:                    # if least significant bit is 1
                R = self._add_aff(R, Q)  # then add current Q
            m = m >> 1                   # remove the bit just accounted for
            Q = self._add_aff(Q, Q)      # double Q for next step
        return R

    # methods using _p: they would become functions if _p goes public

    def opposite(self, Q: Point) -> Point:
        """Return the opposite point on the curve.

        The input point must be on the curve.
        """

        self.require_on_curve(Q)
        # % self._p is required to account for infinity point, i.e. Q[1]==0
        return Q[0], (self._p - Q[1]) % self._p

    def _aff_from_jac(self, Q: _JacPoint) -> Point:
        # point is assumed to be on curve
        if Q[2] == 0:  # Infinity point in Jacobian coordinates
            return 1, 0
        else:
            Z2 = Q[2]*Q[2]
            x = (Q[0]*mod_inv(Z2, self._p)) % self._p
            y = (Q[1]*mod_inv(Z2*Q[2], self._p)) % self._p
            return x, y

    # methods using _a, _b, _p

    def add(self, Q1: Point, Q2: Point) -> Point:
        """Return the sum of two points.

        The input points must be on the curve.
        """

        self.require_on_curve(Q1)
        self.require_on_curve(Q2)
        # no Jacobian coordinates here as _aff_from_jac would cost 2 mod_inv
        # while _add_aff costs only one mod_inv
        return self._add_aff(Q1, Q2)

    def _add_jac(self, Q: _JacPoint, R: _JacPoint) -> _JacPoint:
        # points are assumed to be on curve

        if Q[2] == 0:  # Infinity point in Jacobian coordinates
            return R
        if R[2] == 0:  # Infinity point in Jacobian coordinates
            return Q

        RZ2 = R[2] * R[2]
        RZ3 = RZ2 * R[2]
        QZ2 = Q[2] * Q[2]
        QZ3 = QZ2 * Q[2]
        if Q[0]*RZ2 % self._p == R[0]*QZ2 % self._p:      # same affine x
            if Q[1]*RZ3 % self._p == R[1]*QZ3 % self._p:  # point doubling
                QY2 = Q[1]*Q[1]
                W = (3*Q[0]*Q[0] + self._a*QZ2*QZ2) % self._p
                V = (4*Q[0]*QY2) % self._p
                X = (W*W - 2*V) % self._p
                Y = (W*(V - X) - 8*QY2*QY2) % self._p
                Z = (2*Q[1]*Q[2]) % self._p
                return X, Y, Z
            else:                                         # opposite points
                return 1, 1, 0
        else:
            T = (Q[1]*RZ3) % self._p
            U = (R[1]*QZ3) % self._p
            W = (U - T) % self._p

            M = (Q[0]*RZ2) % self._p
            N = (R[0]*QZ2) % self._p
            V = (N - M) % self._p

            V2 = V * V
            V3 = V2 * V
            MV2 = M * V2
            X = (W*W - V3 - 2*MV2) % self._p
            Y = (W*(MV2 - X) - T*V3) % self._p
            Z = (V*Q[2]*R[2]) % self._p
            return X, Y, Z

    def _add_aff(self, Q: Point, R: Point) -> Point:
        # points are assumed to be on curve
        if R[1] == 0:  # Infinity point in affine coordinates
            return Q
        if Q[1] == 0:  # Infinity point in affine coordinates
            return R

        if R[0] == Q[0]:
            if R[1] == Q[1]:  # point doubling
                lam = (3 * Q[0] * Q[0] + self._a) * mod_inv(2 * Q[1], self._p)
                lam %= self._p
            else:             # opposite points
                return 1, 0
        else:
            lam = ((R[1]-Q[1]) * mod_inv(R[0]-Q[0], self._p)) % self._p
        x = (lam * lam - Q[0] - R[0]) % self._p
        y = (lam * (Q[0] - x) - Q[1]) % self._p
        return x, y

    def _y2(self, x: int) -> int:
        # skipping a crucial check here:
        # if sqrt(y*y) does not exist, then x is not valid.
        # This is a good reason to keep this method private
        return ((x*x + self._a)*x + self._b) % self._p

    def y(self, x: int) -> int:
        """Return the y coordinate from x, as in (x, y)."""
        if not 0 <= x < self._p:
            raise ValueError(f"x-coordinate {hex(x)} not in [0, p-1]")
        y2 = self._y2(x)
        # mod_sqrt will raise a ValueError if root does not exist
        return mod_sqrt(y2, self._p)

    def require_on_curve(self, Q: Point) -> None:
        """Raise an Error if the input point is not on the curve."""
        if not self.is_on_curve(Q):
            raise ValueError("Point not on curve")

    def is_on_curve(self, Q: Point) -> bool:
        """Return True if the input point is on the curve."""
        if len(Q) != 2:
            raise ValueError("Point must be a tuple[int, int]")
        if Q[1] == 0:  # Infinity point in affine coordinates
            return True
        if not 0 < Q[1] < self._p:  # y cannot be zero
            raise ValueError(f"y-coordinate {hex(Q[1])} not in (0, p)")
        return self._y2(Q[0]) == (Q[1]*Q[1] % self._p)

    # break the y simmetry: even/odd, low/high, or quadratic residue criteria

    def y_odd(self, x: int, odd1even0: int = 1) -> int:
        """Return the odd (even) y coordinate from x, as in (x, y)."""
        if odd1even0 not in (0, 1):
            raise ValueError("odd1even0 must be bool or 1/0")
        root = self.y(x)
        # switch even/odd root as needed (XORing the conditions)
        return root if root % 2 == odd1even0 else self._p - root

    def y_low(self, x: int, low1high0: int = 1) -> int:
        """Return the low (high) y coordinate from x, as in (x, y)."""
        if low1high0 not in (0, 1):
            raise ValueError("low1high0 must be bool or 1/0")
        root = self.y(x)
        # switch low/high root as needed (XORing the conditions)
        return root if (self._p//2 >= root) == low1high0 else self._p - root

    def y_quadratic_residue(self, x: int, quad_res: int = 1) -> int:
        """Return the quadratic residue y from x, as in (x, y)."""
        if quad_res not in (0, 1):
            raise ValueError("quad_res must be bool or 1/0")
        if not self.pIsThreeModFour:
            raise ValueError("this method works only when p = 3 (mod 4)")
        root = self.y(x)
        # switch to quadratic residue root as needed
        legendre = legendre_symbol(root, self._p)
        return root if legendre == quad_res else self._p - root



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Strict ASN.1 DER format for ECDSA signature representation.
   
    The original Bitcoin implementation used OpenSSL to verify
    ECDSA signatures in ASN.1 DER representation.
    However, OpenSSL does not do strict validation and
    as long as the signature is not horribly malformed it is accepted.
    E.g. extra padding is ignored and this changes the transaction hash value,
    leading to transaction malleability.
    This was fixed by BIP66, activated on block 363,724.

    source: https://github.com/bitcoin/bips/blob/master/bip-0066.mediawiki

    BIP66 mandates a strict DER format:

    Format: 0x30 [total-length] 0x02 [R-length] [R] 0x02 [S-length] [S] [sighash]

    * total-length: 1-byte length descriptor of everything that follows,
      excluding the sighash byte.
    * R-length: 1-byte length descriptor of the R value that follows.
    * R: arbitrary-length big-endian R value. It must use the shortest
      possible encoding for a positive integers (which means no null bytes at
      the start, except a single one when the next byte has its highest bit set
      to avoid being interpreted as a negative number).
    * S-length: 1-byte length descriptor of the S value that follows.
    * S: arbitrary-length big-endian S value. Same rules of the R value apply.
    * sighash: 1-byte value indicating what data is hashed (not part of the DER
      signature)
"""

from typing import Tuple

from .curve import Curve
from .curves import secp256k1
from .dsa import ECDS, _check_sig

sighash_all = b'\x01'
sighash_none = b'\x02'
sighash_single = b'\x03'
sighash_all_anyonecanpay = b'\x81'
sighash_none_anyonecanpay = b'\x82'
sighash_single_anyonecanpay = b'\x83'


def _bytes_from_scalar(scalar: int) -> bytes:
    # scalar is assumed to be in [1, n-1]
    elen = scalar.bit_length()
    esize = elen // 8 + 1  # not a bug: 'highest bit set' padding included here
    n_bytes = scalar.to_bytes(esize, byteorder='big')
    return n_bytes


def _serialize_scalar(scalar: int) -> bytes:
    # scalar is assumed to be in [1, n-1]
    x = _bytes_from_scalar(scalar)
    xsize = len(x).to_bytes(1, byteorder='big')
    return b'\x02' + xsize + x


def serialize(sig: ECDS, sighash: bytes = sighash_all, ec: Curve = secp256k1) -> bytes:
    """Serialize an ECDSA signature in strict ASN.1 DER representation."""

    if len(sighash) > 1:
        raise ValueError(f"sighash size {len(sighash)} > 1")

    # check that it is a valid signature for the given Curve
    _check_sig(sig, ec)

    enc = _serialize_scalar(sig[0])
    enc += _serialize_scalar(sig[1])
    return b'\x30' + len(enc).to_bytes(1, byteorder='big') + enc + sighash


def deserialize(sig: bytes, ec: Curve = secp256k1) -> Tuple[ECDS, bytes]:
    """Deserialize a strict ASN.1 DER representation of an ECDSA signature."""

    maxsize = ec.nsize * 2 + 7
    sigsize = len(sig)
    if not 8 < sigsize <= maxsize:
        errmsg = f"DER signature size ({sigsize}) must be in "
        errmsg += f"[9, {maxsize}]"
        raise ValueError(errmsg)

    if sig[0] != 0x30:
        raise ValueError("DER signature must be of type 0x30 (compound)")

    # sigsize checks
    if sig[1] + 3 != sigsize:
        m = "Declared signature size does not match with actual signature size"
        raise ValueError(m)

    sizeR = sig[3]  # size of the r scalar
    if sizeR == 0:
        raise ValueError("Zero-size integers are not allowed for r")

    if 5 + sizeR >= sigsize:
        raise ValueError("Size of the s scalar must be inside the signature")

    sizeS = sig[5 + sizeR]  # size of the s scalar
    if sizeS == 0:
        raise ValueError("Zero-size integers are not allowed for s")

    if sizeR + sizeS + 7 != sigsize:
        raise ValueError("Signature size does not match with size of scalars")

    # scalar r
    if sig[2] != 0x02:
        raise ValueError("r scalar must be an integer")

    if sig[4] & 0x80:
        raise ValueError("Negative numbers are not allowed for r")

    # Null bytes at the start of a scalar are not allowed, unless the
    # scalar would otherwise be interpreted as a negative number
    if sizeR > 1 and sig[4] == 0x00 and not (sig[5] & 0x80):
        raise ValueError("Invalid null bytes at the start of r")

    r = int.from_bytes(sig[4:4 + sizeR], byteorder='big')

    # scalar s (offset=2+sizeR with respect to r)
    if sig[sizeR + 4] != 0x02:
        raise ValueError("s scalar must be an integer")

    if sig[sizeR + 6] & 0x80:
        raise ValueError("Negative numbers are not allowed for s")

    # Null bytes at the start of a scalar are not allowed, unless the
    # scalar would otherwise be interpreted as a negative number
    if sizeS > 1 and sig[sizeR + 6] == 0x00 and not (sig[sizeR + 7] & 0x80):
        raise ValueError("Invalid null bytes at the start of s")

    s = int.from_bytes(sig[6 + sizeR:6 + sizeR + sizeS], byteorder='big')

    # checks that the signature is valid for the given Curve
    _check_sig((r, s), ec)
    return (r, s), sig[sigsize - 1:]



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Functions for entropy conversion from/to mnemonic sentence.

Entropy must be represented as binary 0/1 string.

Warning: these functions are not meant for end-users which are
better served by the bip39 and electrum module functions.
"""

import math
from hashlib import pbkdf2_hmac
from typing import List

from .entropy import BinStr
from .wordlists import _wordlists

Mnemonic = str


def _indexes_from_entropy(entropy: BinStr, lang: str) -> List[int]:
    """Return the word-list indexes for a given binary 0/1 string entropy.

    Return the list of integer indexes into a language word-list
    for a given entropy.

    Entropy must be represented as binary 0/1 string; leading zeros
    are not considered redundant padding.
    """

    bits = len(entropy)
    int_entropy = int(entropy, 2)
    n = _wordlists.language_length(lang)
    indexes = []
    while int_entropy:
        int_entropy, index = divmod(int_entropy, n)
        indexes.append(index)

    # do not lose leading zeros entropy
    bpw = _wordlists.bits_per_word(lang)
    nwords = math.ceil(bits/bpw)
    while len(indexes) < nwords:
        indexes.append(0)

    return list(reversed(indexes))


def _mnemonic_from_indexes(indexes: List[int], lang: str) -> Mnemonic:
    """Return the mnemonic from a list of word-list indexes.

    Return the mnemonic from a list of integer indexes into
    a given language word-list.
    """

    words = []
    wordlist = _wordlists.wordlist(lang)
    for i in indexes:
        word = wordlist[i]
        words.append(word)
    return ' '.join(words)


def _indexes_from_mnemonic(mnemonic: Mnemonic, lang: str) -> List[int]:
    """Return the word-list indexes for a given mnemonic.

    Return the list of integer indexes into a language word-list
    for a given mnemonic.
    """

    words = mnemonic.split()
    wordlist = _wordlists.wordlist(lang)
    indexes = [wordlist.index(w) for w in words]
    return indexes


def _entropy_from_indexes(indexes: List[int], lang: str) -> BinStr:
    """Return the entropy from a list of word-list indexes.

    Return the entropy from a list of integer indexes into
    a given language word-list.
    """

    n = _wordlists.language_length(lang)
    entropy = 0
    for i in indexes:
        entropy = entropy*n + i

    binentropy = bin(entropy)[2:]    # remove '0b'

    # do not lose leading zeros entropy
    bpw = _wordlists.bits_per_word(lang)
    bits = len(indexes)*bpw
    binentropy = binentropy.zfill(bits)

    return binentropy



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Elliptic Curve Schnorr Signature Algorithm (ECSSA).

Implementation according to bip-schnorr:

https://github.com/sipa/bips/blob/bip-schnorr/bip-schnorr.mediawiki.
"""

import heapq
import random
from typing import List, Optional, Sequence, Tuple
from hashlib import sha256

from .curve import Curve, Point, _JacPoint
from .curves import secp256k1
from .curvemult import (_double_mult, _jac_from_aff, _mult_jac,
                        _multi_mult, double_mult, mult)
from .numbertheory import legendre_symbol, mod_inv
from .rfc6979 import rfc6979
from .utils import HashF, int_from_bits, octets_from_int, octets_from_point

ECSS = Tuple[int, int]  # Tuple[field element, scalar]


def _ensure_msg_size(msg: bytes, hf: HashF = sha256) -> None:
    if len(msg) != hf().digest_size:
        errmsg = f'message of wrong size: {len(msg)}'
        errmsg += f' instead of {hf().digest_size} bytes'
        raise ValueError(errmsg)


def _e(r: int, P: Point, mhd: bytes, ec: Curve = secp256k1, hf: HashF = sha256) -> int:
    # Let e = int(hf(bytes(x(R)) || bytes(dG) || mhd)) mod n.
    h = hf()
    h.update(octets_from_int(r, ec.psize))
    h.update(octets_from_point(P, True, ec))
    h.update(mhd)
    e = int_from_bits(h.digest(), ec)
    return e


def sign(mhd: bytes, d: int, k: Optional[int] = None,
         ec: Curve = secp256k1, hf: HashF = sha256) -> ECSS:
    """ECSSA signing operation according to bip-schnorr.

    This signature scheme supports only 32-byte messages.
    Differently from ECDSA, the 32-byte message can be a
    digest of other messages, but it does not need to.
    """

    # the bitcoin proposed standard is only valid for curves
    # whose prime p = 3 % 4
    if not ec.pIsThreeModFour:
        errmsg = 'curve prime p must be equal to 3 (mod 4)'
        raise ValueError(errmsg)

    # The message mhd: a 32-byte array
    _ensure_msg_size(mhd, hf)

    # The secret key d: an integer in the range 1..n-1.
    if not 0 < d < ec.n:
        raise ValueError(f"private key {hex(d)} not in [1, n-1]")
    P = mult(d, ec.G, ec)

    # Fail if k' = 0.
    if k is None:
        k = rfc6979(mhd, d, ec, hf)
    if not 0 < k < ec.n:
        raise ValueError(f"ephemeral key {hex(k)} not in [1, n-1]")

    # Let R = k'G.
    RJ = _mult_jac(k, ec.GJ, ec)

    # break the simmetry: any criteria might have been used,
    # jacobi is the proposed bitcoin standard
    # Let k = k' if jacobi(y(R)) = 1, otherwise let k = n - k'.
    if legendre_symbol(RJ[1]*RJ[2] % ec._p, ec._p) != 1:
        k = ec.n - k

    Z2 = RJ[2]*RJ[2]
    r = (RJ[0]*mod_inv(Z2, ec._p)) % ec._p

    # Let e = int(hf(bytes(x(R)) || bytes(dG) || mhd)) mod n.
    e = _e(r, P, mhd, ec, hf)

    s = (k + e*d) % ec.n  # s=0 is ok: in verification there is no inverse of s
    # The signature is bytes(x(R) || bytes((k + ed) mod n)).
    return r, s


def verify(mhd: bytes, P: Point, sig: ECSS,
           ec: Curve = secp256k1, hf: HashF = sha256) -> bool:
    """ECSSA signature verification according to bip-schnorr."""

    # try/except wrapper for the Errors raised by _verify
    try:
        return _verify(mhd, P, sig, ec, hf)
    except Exception:
        return False


def _verify(mhd: bytes, P: Point, sig: ECSS,
            ec: Curve = secp256k1, hf: HashF = sha256) -> bool:
    # Private function for test/dev purposes
    # It raises Errors, while verify should always return True or False

    # the bitcoin proposed standard is only valid for curves
    # whose prime p = 3 % 4
    if not ec.pIsThreeModFour:
        errmsg = 'curve prime p must be equal to 3 (mod 4)'
        raise ValueError(errmsg)

    # Let r = int(sig[ 0:32]).
    # Let s = int(sig[32:64]); fail if s is not [0, n-1].
    _check_sig(sig, ec)

    # The message mhd: a 32-byte array
    _ensure_msg_size(mhd, hf)

    # Let P = point(pk); fail if point(pk) fails.
    ec.require_on_curve(P)
    if P[1] == 0:
        raise ValueError("public key is infinite")

    # Let e = int(hf(bytes(r) || bytes(P) || mhd)) mod n.
    e = _e(sig[0], P, mhd, ec, hf)

    # Let R = sG - eP.
    # in Jacobian coordinates
    R = _double_mult(-e, (P[0], P[1], 1), sig[1], ec.GJ, ec)

    # Fail if infinite(R).
    if R[2] == 0:
        raise ValueError("sG - eP is infinite")

    # Fail if jacobi(R.y) ≠ 1.
    if legendre_symbol(R[1]*R[2] % ec._p, ec._p) != 1:
        raise ValueError("(sG - eP).y is not a quadratic residue")

    # Fail if R.x ≠ r.
    return R[0] == (R[2]*R[2]*sig[0] % ec._p)


def _pubkey_recovery(e: int, sig: ECSS,
                     ec: Curve = secp256k1, hf: HashF = sha256) -> Point:
    # Private function provided for testing purposes only.
    # TODO: use _double_mult instead of double_mult

    _check_sig(sig, ec)

    K = sig[0], ec.y_quadratic_residue(sig[0], True)
    # FIXME: y_quadratic_residue in Jacobian coordinates?

    if e == 0:
        raise ValueError("invalid (zero) challenge e")
    e1 = mod_inv(e, ec.n)
    P = double_mult(-e1, K, e1*sig[1], ec.G, ec)
    assert P[1] != 0, "how did you do that?!?"
    return P


def _check_sig(sig: ECSS, ec: Curve = secp256k1) -> None:
    # check that the SSA signature is correct
    # and return the signature itself

    # A signature sig: a 64-byte array.
    if len(sig) != 2:
        mhd = f"invalid length {len(sig)} for ECSSA signature"
        raise TypeError(mhd)

    # Let r = int(sig[ 0:32]); fail if r is not a valid x-coordinate.
    ec.y(sig[0])

    # Let s = int(sig[32:64]); fail if s is not [0, n-1].
    if not 0 <= sig[1] < ec.n:
        raise ValueError(f"s ({hex(sig[1])}) not in [0, n-1]")



def batch_verify(ms: Sequence[bytes], P: Sequence[Point], sig: Sequence[ECSS],
                 ec: Curve = secp256k1, hf: HashF = sha256) -> bool:
    """ECSSA batch signature verification according to bip-schnorr."""

    # try/except wrapper for the Errors raised by _batch_verify
    try:
        return _batch_verify(ms, P, sig, ec, hf)
    except Exception:
        return False


def _batch_verify(ms: Sequence[bytes], P: Sequence[Point], sig: Sequence[ECSS],
                  ec: Curve = secp256k1, hf: HashF = sha256) -> bool:

    # the bitcoin proposed standard is only valid for curves
    # whose prime p = 3 % 4
    if not ec.pIsThreeModFour:
        errmsg = 'curve prime p must be equal to 3 (mod 4)'
        raise ValueError(errmsg)

    batch_size = len(P)
    if len(ms) != batch_size:
        errMsg = f"mismatch between number of pubkeys ({batch_size}) "
        errMsg += f"and number of messages ({len(ms)})"
        raise ValueError(errMsg)
    if len(sig) != batch_size:
        errMsg = f"mismatch between number of pubkeys ({batch_size}) "
        errMsg += f"and number of signatures ({len(sig)})"
        raise ValueError(errMsg)

    if batch_size == 1:
        return _verify(ms[0], P[0], sig[0], ec, hf)

    t = 0
    scalars: List[int] = list()
    points: List[_JacPoint] = list()
    for i in range(batch_size):
        _check_sig(sig[i], ec)
        r, s = sig[i]
        _ensure_msg_size(ms[i], hf)
        ec.require_on_curve(P[i])
        e = _e(r, P[i], ms[i], ec, hf)
        # raises an error if y does not exist
        # no need to check for quadratic residue
        y = ec.y(r)

        # a in [1, n-1]
        # deterministically generated using a CSPRNG seeded by a
        # cryptographic hash (e.g., SHA256) of all inputs of the
        # algorithm, or randomly generated independently for each
        # run of the batch verification algorithm
        a = (1 if i == 0 else (1+random.getrandbits(ec.nlen)) % ec.n)
        scalars.append(a)
        points.append(_jac_from_aff((r, y)))
        scalars.append(a * e % ec.n)
        points.append(_jac_from_aff(P[i]))
        t += a * s

    TJ = _mult_jac(t, ec.GJ, ec)
    RHSJ = _multi_mult(scalars, points, ec)

    # return T == RHS, checked in Jacobian coordinates
    RHSZ2 = RHSJ[2] * RHSJ[2]
    TZ2 = TJ[2] * TJ[2]
    if (TJ[0] * RHSZ2) % ec._p != (RHSJ[0] * TZ2) % ec._p:
        return False

    return (TJ[1] * RHSZ2 * RHSJ[2]) % ec._p == (RHSJ[1] * TZ2 * TJ[2]) % ec._p



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Elliptic Curve Digital Signature Algorithm (ECDSA).

   Implementation according to SEC 1 v.2:

   http://www.secg.org/sec1-v2.pdf

   specialized with bitcoin canonical 'low-s' encoding.
"""

from typing import Tuple, List, Optional
from hashlib import sha256

from .numbertheory import mod_inv
from .utils import int_from_bits, HashF
from .curve import Curve, Point
from .curves import secp256k1
from .curvemult import _mult_jac, _double_mult, double_mult
from .rfc6979 import _rfc6979

ECDS = Tuple[int, int]  # Tuple[scalar, scalar]


def sign(m: bytes, q: int, k: Optional[int] = None,
         ec: Curve = secp256k1, hf: HashF = sha256) -> ECDS:
    """ECDSA signature according to SEC 1 v.2 with canonical low-s encoding.

    The message m is first processed by hf, yielding the value

        mhd = hf(m),

    a sequence of bits of length *hlen*.

    Normally, hf is chosen such that its output length *hlen* is
    roughly equal to *nlen*, the bit-length of the group order *n*,
    since the overall security of the signature scheme will depend on
    the smallest of *hlen* and *nlen*; however, the ECDSA standard
    supports all combinations of *hlen* and *nlen*.

    See https://tools.ietf.org/html/rfc6979#section-3.2
    """

    # Steps numbering follows SEC 1 v.2 section 4.1.3

    mhd = hf(m).digest()                          # 4
    # mhd is transformed into an integer modulo ec.n using int_from_bits:
    c = int_from_bits(mhd, ec)                    # 5

    # The secret key q: an integer in the range 1..n-1.
    # SEC 1 v.2 section 3.2.1
    if not 0 < q < ec.n:
        raise ValueError(f"private key {hex(q)} not in [1, n-1]")

    if k is None:
        k = _rfc6979(c, q, ec, hf)                # 1
    if not 0 < k < ec.n:
        raise ValueError(f"ephemeral key {hex(k)} not in [1, n-1]")

    # second part delegated to helper function
    return _sign(c, q, k, ec)


def _sign(c: int, q: int, k: int, ec: Curve = secp256k1) -> ECDS:
    # Private function for test/dev purposes
    # it is assumed that q, k, and c are in [1, n-1]

    # Steps numbering follows SEC 1 v.2 section 4.1.3

    RJ = _mult_jac(k, ec.GJ, ec)                  # 1

    Rx = (RJ[0]*mod_inv(RJ[2]*RJ[2], ec._p)) % ec._p
    r = Rx % ec.n                                 # 2, 3
    if r == 0:  # r≠0 required as it multiplies the public key
        raise ValueError("r = 0, failed to sign")

    s = mod_inv(k, ec.n) * (c + r*q) % ec.n       # 6
    if s == 0:  # s≠0 required as verify will need the inverse of s
        raise ValueError("s = 0, failed to sign")

    # bitcoin canonical 'low-s' encoding for ECDSA signatures
    # it removes signature malleability as cause of transaction malleability
    # see https://github.com/bitcoin/bitcoin/pull/6769
    if s > ec.n / 2:
        s = ec.n - s

    return r, s


def verify(m: bytes, P: Point, sig: ECDS,
           ec: Curve = secp256k1, hf: HashF = sha256) -> bool:
    """ECDSA signature verification (SEC 1 v.2 section 4.1.4)."""

    # try/except wrapper for the Errors raised by _verify
    try:
        return _verify(m, P, sig, ec, hf)
    except Exception:
        return False


def _verify(m: bytes, P: Point, sig: ECDS,
            ec: Curve = secp256k1, hf: HashF = sha256) -> bool:
    # Private function for test/dev purposes
    # It raises Errors, while verify should always return True or False

    # The message digest mhd: a 32-byte array
    mhd = hf(m).digest()                                 # 2
    c = int_from_bits(mhd, ec)                           # 3

    # second part delegated to helper function
    return _verhlp(c, P, sig, ec)


def _verhlp(c: int, P: Point, sig: ECDS, ec: Curve = secp256k1) -> bool:
    # Private function for test/dev purposes

    # Fail if r is not [1, n-1]
    # Fail if s is not [1, n-1]
    _check_sig(sig, ec)                                  # 1

    # Let P = point(pk); fail if point(pk) fails.
    ec.require_on_curve(P)
    if P[1] == 0:
        raise ValueError("public key is infinite")

    w = mod_inv(sig[1], ec.n)
    u = c*w
    v = sig[0]*w                                         # 4
    # Let R = u*G + v*P.
    RJ = _double_mult(v, (P[0], P[1], 1), u, ec.GJ, ec)  # 5

    # Fail if infinite(R).
    assert RJ[2] != 0, "how did you do that?!?"          # 5

    Rx = (RJ[0]*mod_inv(RJ[2]*RJ[2], ec._p)) % ec._p
    x = Rx % ec.n                                        # 6, 7
    # Fail if r ≠ x(R) %n.
    return sig[0] == x                                   # 8


def pubkey_recovery(m: bytes, sig: ECDS,
                    ec: Curve = secp256k1, hf: HashF = sha256) -> List[Point]:
    """ECDSA public key recovery (SEC 1 v.2 section 4.1.6).

    See also https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106#18106    
    """

    # The message digest mhd: a 32-byte array
    mhd = hf(m).digest()                                  # 1.5
    c = int_from_bits(mhd, ec)                            # 1.5

    return _pubkey_recovery(c, sig, ec)


def _pubkey_recovery(c: int, sig: ECDS, ec: Curve = secp256k1) -> List[Point]:
    # Private function provided for testing purposes only.
    # TODO: use _double_mult instead of double_mult

    _check_sig(sig, ec)

    # precomputations
    r1 = mod_inv(sig[0], ec.n)
    r1s = r1*sig[1]
    r1e = -r1*c
    keys: List[Point] = list()
    # r = R[0] % ec.n
    # if ec.n < R[0] < ec._p (probable when cofactor ec.h > 1)
    # then both x=r and x=r+ec.n must be tested
    for j in range(ec.h):                                 # 1
        x = (sig[0] + j*ec.n) % ec._p                     # 1.1
        try:
            # even root first for bitcoin message signing compatibility
            R = x, ec.y_odd(x, False)                     # 1.2, 1.3, and 1.4
            # 1.5 has been performed in the pubkey_recovery calling function
            Q1 = double_mult(r1s, R, r1e, ec.G, ec)       # 1.6.1
            if Q1[1] != 0 and _verhlp(c, Q1, sig, ec):    # 1.6.2
                keys.append(Q1)
            R = ec.opposite(R)                            # 1.6.3
            Q2 = double_mult(r1s, R, r1e, ec.G, ec)
            if Q2[1] != 0 and _verhlp(c, Q2, sig, ec):    # 1.6.2
                keys.append(Q2)                           # 1.6.2
        except Exception:  # R is not a curve point
            pass
    return keys


def _check_sig(sig: ECDS, ec: Curve = secp256k1) -> None:
    # check that the DSA signature is correct
    # and return the signature itself

    if len(sig) != 2:
        errMsg = f"invalid length {len(sig)} for ECDSA signature"
        raise TypeError(errMsg)

    # Fail if r is not [1, n-1]
    if not 0 < sig[0] < ec.n:
        raise ValueError(f"r ({hex(sig[0])}) not in [1, n-1]")

    # Fail if s is not [1, n-1]
    if not 0 < sig[1] < ec.n:
        raise ValueError(f"s ({hex(sig[1])}) not in [1, n-1]")
